from pathlib import Path
import os, time
from treerec.primitives import *
from treerec.hist_events import record_entry
import hashlib
import uuid



def hash_file(pathlike):
    """
    Given a Path object, return the sha256 hash of its contents.
    By convention, a hash of -1 indicates that the file is a directory,
      a hash of -2 indicates a lack of permission to hash the file,
      and a hash of 0 indicates that the hash was not taken on a file.
    """
    if pathlike.is_dir():
        return -1
    try:
        with open(pathlike, 'rb') as f:
            return hashlib.sha256(f.read()).hexdigest()
    except PermissionError:
        return -2
    except:
        return 0


def fileobj_get_owner(pathlike):
    if os.name=='nt':
        from treerec import winutils
        return winutils.get_win_info(pathlike)[0]
    elif os.name=='posix':
        return pathlike.owner()


def fileobj_alphabetize(dictin:dict):
    if dictin.get('dir_list', False):
        dictin['dir_list'] = sorted(dictin['dir_list'], key=lambda x: str.casefold(x['name']))
    if dictin.get('file_list', False):
        dictin['file_list'] = sorted(dictin['file_list'], key=lambda x: str.casefold(x['name']))
       

def create_dict(pathlike, rootname=None, device=None, os_type=None, session=None, objtype='auto', nickname=None, do_hash=True, include_uuid=False):
    try:
        stat = pathlike.stat()
    except FileNotFoundError:
        print("File not found at path " + str(pathlike))
        return False
    objtype = (('dir' if pathlike.is_dir() else 'file') if objtype=='auto' else objtype)
    hash_data = (hash_file(pathlike) if do_hash else 0)
    out = {
        'name' : pathlike.name,
        'root' : rootname,
        'path' : pathlike.parts,
        'type' : objtype,
        'owner' : fileobj_get_owner(pathlike),
        'device' : device,
        'os_type' : (os_type if os_type else os.name),
        'ctime' : int(stat.st_ctime),
        'mtime' : int(stat.st_mtime),
        'atime' : int(stat.st_atime),
        'hash' : hash_data,
        'notes' : [],
        'ties' : [],
        'history' : []
    }
    out['history'].append(record_entry(out, session=session, localpath=str(pathlike)))
    if include_uuid:
        out['uuid'] = uuid.uuid4().hex
    if objtype in ['rootdir', 'dir']:
        contents_list = [pathlike/entry for entry in os.listdir(pathlike)]
        out['dir_list'] = [{'name' : entry.name, 'type' : 'unrecorded'} for entry in contents_list if Path(entry).is_dir()]
        out['file_list'] = [{'name' : entry.name, 'type' : 'unrecorded'} for entry in contents_list if not Path(entry).is_dir()]
        fileobj_alphabetize(out) # I think the sorting is unnecessary, but I'll do it anyway for good measure
    if nickname:
        out['nickname'] = nickname
    return out


def create_dict_pure(
        objname:str=None, 
        objtype:str=None,
        objpath:str=None,
        owner:str=None,
        device:str=None,
        os_type:str=None,
        ctime:int=None,
        mtime:int=None,
        atime:int=None,
        shortcut_destination=None,
        nickname:str=None,
        include_uuid:bool=False):
    out = {
        'name' : objname,
        'path' : objpath,
        'type' : objtype,
        'owner' : owner,
        'device' : device,
        'os_type' : os_type,
        'ctime': ctime,
        'mtime' : mtime,
        'atime' : atime,
        'hash' : 0,
        'notes' : [],
        'ties' : [],
        'history' : []
    }
    if include_uuid:
        out['uuid'] = uuid.uuid4().hex
    if objtype in DIR_LIST_TYPES:
        out['dir_list'] = []
        out['file_list'] = []
    if objtype=='shortcut':
        out['shortcut_destination'] = shortcut_destination
    if nickname!=None:
        out['nickname'] = nickname
    return out


def get_position_in_parent(dictin:dict, nickname_first:bool=False):
    """
    Returns the index and list of dictin in its parent dict.
    """
    parent = dictin.get(PARENT, False)
    if not parent:
        print("Error finding parent dictionary.")
        return False, False
    if nickname_first:
        name_priority = ['nickname', 'name']
    else:
        name_priority = ['name', 'nickname']
    for nametype in name_priority:
        for x_list in ['dir_list', 'file_list']:
            for idx, entry in enumerate(parent[x_list]):
                if entry.get(nametype) and entry.get(nametype)==dictin.get(nametype):
                    return idx, x_list
    print("Entry name not found in parent dictionary.")
    return False, False


def find_current_root(dirdict):
    if dirdict['type'] == DEVINFO:
        return dirdict
    curdir = dirdict
    parentdir = dirdict[('parent',)]
    while parentdir['type'] != DEVINFO:
        curdir = parentdir
        parentdir = curdir[('parent',)]
    return curdir


def get_tree_id(dirdict):
    """
    Given a directory dict, return the ID of the tree it is in.
    """
    root = find_current_root(dirdict)
    return root[('parent',)]['uuid']


def get_recorded_session_name(dictin):
    for event in dictin['history']:
        if event['event']['type']=='record':
            return event['session']


def print_basic_info(dictin):

    def print_thing(label, key, default='N/A'):
        dictin_get = dictin.get(key, None)
        print(label + ': ' + (str(dictin_get) if dictin_get!=None else default))

    print("Name: " + dictin['name'])
    print("Type: " + dictin['type'])
    if dictin['type']=='unrecorded':
        return
    first_modified_session = get_recorded_session_name(dictin)
    last_modified_session = dictin['history'][-1]['session']
    print_thing('Path', 'path')
    if dictin.get('localpath', None):
        print_thing('Local Path', 'localpath')
    print_thing('Owner', 'owner')
    print_thing('Device', 'device')
    print_thing('OS type', 'os_type')
    print('Recorded during: Session "', first_modified_session,'".')
    print('Last modified during: Session "', last_modified_session,'".')
    print_thing('Creation time', 'ctime')
    print_thing('Modification time', 'mtime')
    print_thing('Access time', 'atime')
    print_thing('Notes', 'notes', default='None')
    print_thing('Ties', 'ties', default='None')
    print_thing('History', 'history', default='None')


def print_dict_history(dictin):
    for entry in dictin.get('history', None):
        event_type = entry['event']['type']
        print('>> ' + event_type + " // " + time.ctime(entry['time']))
        print('   Session: ' + entry.get('session', 'None'))
        if event_type=='tie':
            print('   -------------')
            print('   Name: ' + entry['event']['name'])
            print('   Device: ' + entry['event']['device'])
        print('  ')


def dict_add_note(dictin, note_text='', session=None):
    if note_text=='':
        return False
    dictin['notes'].append({
        'time' : time.time(),
        'session' : session,
        'text' : note_text
    })
    return True


def print_dict_notes(dictin):
    if len(dictin.get('notes', []))==0:
        print(' [no notes]')
        return
    for idx, entry in enumerate(dictin.get('notes', [])):
        print('>> Note ' + str(idx) + ' // ' + time.ctime(entry['time']))
        print('   Session: ' + entry.get('session', 'None'))
        print(entry['text'])
        print('-------------------------')



from treerec.primitives import DEVINFO

def find_current_root(dirdict):
    if dirdict['type'] == DEVINFO:
        return dirdict
    curdir = dirdict
    parentdir = dirdict[('parent',)]
    while parentdir['type'] != DEVINFO:
        curdir = parentdir
        parentdir = curdir[('parent',)]
    return curdir


def get_tree_id(dirdict):
    """
    Given a directory dict, return the ID of the tree it is in.
    """
    root = find_current_root(dirdict)
    return root[('parent',)]['uuid']



import treerec

T = treerec.TreeObj_Cursor()
T.execute('session K')
T.execute('create new Dev .')
T.execute('recroot -n T -p / treerec')
T.execute('cd $ "C:/Program Files (x86)/Steam/steamapps/common"')
T.execute('recroot -n steam .')

T.mainloop()


from treerec.primitives import *
from treerec.pathtuputils import *
from treerec.dictutils import *


def calculate_localpath(dictin:dict):
    if dictin['type']=='unrecorded':
        return False
    if dictin.get('nickname', False):
        return dictin['localpath']
    elif len(dictin['path'])==0:
        print("Error: no path for object " + dictin['name'])
        return False
    pathparts = [dictin['name']]
    cdir = dictin
    while not cdir[PARENT].get('nickname', False):
        cdir = cdir[PARENT]
        pathparts = [cdir['name']] + pathparts
    cdir = cdir[PARENT]
    # now, cdir is the root directory
    return (*cdir['localpath'], *pathparts)


def find_in_dir(dirdict, objname, nickname_first=False):
    """
    Given a directory dict and an objname, return the child dict with the name or nickname of objname,
        along with the list it was found in and the index of the entry in that list.
    If no entry is found, return False, False, False.
    """
    if nickname_first:
        for idx, entry in enumerate(dirdict.get('dir_list')):
            if entry.get('nickname') == objname:
                return entry, 'dir_list', idx
    for idx, entry in enumerate(dirdict.get('dir_list')):
        if entry['name'] == objname:
            return entry, 'dir_list', idx
    for idx, entry in enumerate(dirdict.get('file_list')):
        if entry['name'] == objname:
            return entry, 'file_list', idx
    for idx, entry in enumerate(dirdict.get('dir_list')):
        if entry.get('nickname') == objname:
            return entry, 'dir_list', idx
    return False, False, False


def locate_from_dir(dirdict, objname, nickname_first=False):
    """
    Same as find_in_dir, but handles the case where objname='.' or '..' to return the current/parent directory
    """
    if objname=='.':
        return dirdict, 'dir_list', None
    if objname=='..':
        return dirdict[PARENT], 'dir_list', None
    return find_in_dir(dirdict, objname, nickname_first=nickname_first)


def replace_entry_in_dir(direc, new_dict, quiet=False):
    """
    Replaces an "unrecorded" child dict with a "recorded" dict.
    """
    old_dict, x_list, idx = find_in_dir(direc, new_dict['name'])
    if old_dict==False and not quiet:
        print(new_dict['name'] + " not found in directory " + direc['name'])
        return False
    if direc[x_list][idx]['type']=='unrecorded':
        direc[x_list][idx] = new_dict
        direc[x_list][idx][PARENT] = direc
        return True
    else:
        if not quiet:
            print("Entry already recorded.")
        return False


def recursive_dir_list_gen(dictin, only_recorded_dirs=False):
    yield from dir_children_generator(dictin, 
                                      types_to_include=['dir_list'], 
                                      only_recorded_entries=only_recorded_dirs)


def dir_children_generator(dictin:dict, 
                           types_to_include:tuple=['file_list', 'dir_list'], 
                           only_recorded_entries:bool=False):
    yield dictin
    for x_list in types_to_include:
        for entry in dictin.get(x_list, []): # responds correctly to files
            if entry['type']!='unrecorded' or not only_recorded_entries:
                yield from dir_children_generator(entry, types_to_include, only_recorded_entries)


def setup_parents(dictin):
    for entry in recursive_dir_list_gen(dictin, only_recorded_dirs=True):
        for sub_entry in entry['dir_list']+entry['file_list']:
            sub_entry[PARENT] = entry
    return dictin


def dir_is_empty(dictin):
    dirs = [entry['type']=='unrecorded' for entry in dictin['dir_list']]
    files = [entry['type']=='unrecorded' for entry in dictin['file_list']]
    return all(dirs + files)


### Code for moving a file


def bare_move_entry(source_dict:dict, destination_dict:dict) -> bool:
    """
    Code for inserting source_dict into the destination_dict, 
    alphabetizing the children,
    and removing the source_dict from its original parent dict.
    """
    if destination_dict['type'] not in DIR_LIST_TYPES:
        return False
    source_parent = source_dict[PARENT]
    entry, x_list, idx = find_in_dir(source_parent, source_dict['name'])
    destination_dict[x_list].append(source_dict)
    fileobj_alphabetize(destination_dict)
    source_parent[x_list].pop(idx)
    return True


def refresh_entry_paths(dictin:dict, new_root_name:str=None):
    """
    Iteratively goes through every child of dictin, and gives it a path
    based on the path of dictin and on its own path from dictin.
    """
    if dictin['type']==DEVINFO:
        return
    path_walk_gen = dir_children_generator(dictin, only_recorded_entries=True)
    if dictin['type']=='rootdir':
        # this means that dirin is a root, and its parent (devinfo) has no path.
        next(path_walk_gen)
    for entry in path_walk_gen:
        entry['path'] = (*entry[PARENT]['path'], entry['name'])
        if not (new_root_name is None):
            entry['root'] = new_root_name


def move_entry(source_dict:dict, destination_dict:dict, new_name:str=None):
    bare_move_entry(source_dict, destination_dict)
    new_root_name = None
    if new_name:
        source_dict['name'] = new_name
        if source_dict['type']=='rootdir':
            new_root_name = source_dict['name']
    refresh_entry_paths(source_dict, new_root_name)



import json, os
from treerec.primitives import *
from treerec.dictutils import *
from treerec.entryutils import time_formatted, tie_files, print_os_file_info
from treerec.treeobj import TreeObj


forbidden_tree_nicknames = ['*', '', ' ', DEFAULT_SYSCHAR, '.', '..', '/', '\\', '^', '^^']

def checkname(allowsys=False, replace=True):
    """
    Decorator method for checking if a Tree name is valid.
    If the Tree name is valid, the method is passed through.
    If the Tree name is None, it replaces it with the current Tree,
        unless replace=False, in which case it prints an error message.
    If the Tree name is invalid, it prints an error message, and passes.
    """
    def checkname_decorator(func):
        def checkname_wrapper(mth, treename=None, *args, **kwargs):
            if treename==None and replace:
                if mth.CTREE==None:
                    print('No trees in current session.')
                    return None
                treename = mth.CTREE
            if treename==mth.SYSCHAR and not allowsys:
                print('Cannot use this method on the system.')
                return None
            if (treename != mth.SYSCHAR) and (mth.treedict.get(treename) == None):
                print('Tree "' + treename + '" does not exist.')
                return None
            return func(mth, treename, *args, **kwargs)
        return checkname_wrapper
    return checkname_decorator


def ls_os(pathstr:str=None, show_mtime:bool=False):
    if not pathstr:
        pathstr = '.'
    pathlike = Path(pathstr).resolve().absolute()
    if not pathlike.exists():
        print('Error locating path ' + str(pathlike))
        return
    try:
        entrylist = os.listdir(pathstr)
    except PermissionError:
        print("Permission denied.")
        return
    except:
        print("Unable to complete request.")
        return
    dir_list = [entry for entry in entrylist if Path(pathlike/entry).is_dir()]
    file_list = [entry for entry in entrylist if not Path(pathlike/entry).is_dir()]
    dir_sorted = sorted(dir_list, key=lambda x: str.casefold(x))
    file_sorted = sorted(file_list, key=lambda x: str.casefold(x))
    for entry in dir_sorted:
        timestr = ('     ' + time_formatted(Path(pathlike/entry).stat().st_mtime) if show_mtime else '')
        print('   [] ' + entry + timestr)
    for entry in file_sorted:
        timestr = ('     ' + time_formatted(Path(pathlike/entry).stat().st_mtime) if show_mtime else '')
        print('      ' + entry + timestr)


class TreeEngine_Base():

    # These two are here to prevent error messages in Jupyter;
    # they are unused in the code
    shape = ()
    __len__ = lambda self: len(self.treedict)

    def __init__(self, session:str=None):
        self.session = session
        self.ignore_names_list = []
        self.treedict = {}
        self.SYSCHAR = DEFAULT_SYSCHAR
        self.CTREE = self.SYSCHAR

    def __getattr__(self, attr):
        """
        Non-MultiTreeHolder methods are passed through to the current tree.
        """
        if self.CTREE == None:
            print('>>> ' + attr + ' <<<')
            print('No tree is currently active.')
            return self.blankfunc
        if self.CTREE == self.SYSCHAR:
            print('Cannot call method or attribute "' + attr + '" on the system.')
            return self.blankfunc
        tree = self.treedict[self.CTREE] #  Without this declaration, the kernal crashes...
        return getattr(tree, attr)

    def blankfunc(self, *args, **kwargs):
        return None


    ### code for setting tree properties
    ### i.e. onsystem, active, savefile, session


    @checkname(allowsys=False, replace=True)
    def set_onsystem(self, treename:str=None, device:str=None):
        self.treedict[treename].set_onsystem(device=device)

    @checkname(allowsys=False, replace=True)
    def set_offsystem(self, treename:str=None):
        self.treedict[treename].set_offsystem()

    @checkname(allowsys=False, replace=True)
    def activate(self, treename:str=None):
        if not self.session:
            print('Set Session name before activating any trees.')
        self.treedict[treename].activate(self.session)

    @checkname(allowsys=False, replace=True)
    def deactivate(self, treename:str=None):
        self.treedict[treename].deactivate()

    @checkname(allowsys=False, replace=True)
    def set_hash_rule(self, treename:str=None, do_hash:bool=True):
        """Sets the hash rule: whether to hash files when recording onto the Tree."""
        self.treedict[treename].set_hash_rule(do_hash)

    @checkname(allowsys=False, replace=True)
    def set_auto_update(self, treename:str=None, action:bool=False):
        TREE = self.treedict[treename]
        TREE.auto_update = bool(action)
        self.show_auto_update(treename)
        if TREE.auto_update and not TREE.onsystem:
            print('Warning: "' + treename + '" is not on-system, and will not update directory contents.')

    @checkname(allowsys=False, replace=True)
    def show_auto_update(self, treename:str=None):
        print('Autoupdate is ' + 
                ('on' if self.treedict[treename].auto_update else 'off') + 
                ' for Tree "' + treename + '".')

    @checkname(allowsys=False, replace=True)
    def set_savefile(self, treename:str, filename:str, quiet:bool=False, force:bool=False):
        self.treedict[treename].set_savefile(filename, quiet, force)

    @checkname(allowsys=False, replace=True)
    def save(self, treename:str=False, force:bool=True, pretty:bool=False):
        """
        Save the Tree to its Savefile.
        If Tree has no unsaved changes, we do not save unless force=True.
        force=True won't save if there is no Savefile set (Tree.savefile_active == False)
          or if the Tree is not active.
        """
        if not self.treedict[treename].savefile_active:
            print('Warning: Tree not saved. "' + treename + '" has no valid save file.')
            return False
        if not self.treedict[treename].has_unsaved_changes and not force:
            print('Tree "' + treename + '" has no unsaved changes, and will not be saved. Override this with the -f flag.')
            return False
        if not self.treedict[treename].is_active():
            print('Tree "' + treename + '" must be active to perform this operation')
            return False
        saved = self.treedict[treename].save(pretty=pretty)
        if saved:
            print('Tree "' + treename + '" saved...')
            return True
        else:
            return False

    def saveall(self, force:bool=True, pretty:bool=False):
        for treename in self.treedict:
            if self.treedict[treename].has_unsaved_changes or force:
                # The extra condition here prevents the "no unsaved changes" response
                # from popping up on the unsaved trees.
                self.save(treename, force=force, pretty=pretty)

    def has_unsaved(self):
        for treename in self.treedict:
            if self.treedict[treename].has_unsaved_changes:
                return True
        return False

    def set_session_name(self, name:str):
        self.session = name
        self.update_tree_session_names()

    def update_tree_session_names(self):
        for tree in self.treedict.values():
            tree.session = self.session

    def check_nickname_is_valid(self, nickname:str=None):
        if not nickname:
            print('Error: Tree Nickname cannot be blank.')
            return False  
        if '/' in nickname or '\\' in nickname:
            print('Error: forbidden Tree Nickname. Nicknames cannot include "/" or "\\".')
            return False
        if nickname in forbidden_tree_nicknames:
            print('Error: forbidden Tree Nickname.')
            return False
        return True

    @checkname(allowsys=False, replace=True)
    def rename(self, oldname: str, newname: str):
        if not self.check_nickname_is_valid(newname):
            return
        if self.treedict.get(newname) != None:
            print('Error: cannot rename. Tree with name "' + newname + '" already exists.')
            return
        self.treedict[newname] = self.treedict.pop(oldname)
        if self.CTREE == oldname:
            self.CTREE = newname
        print('Tree "' + oldname + '" renamed to "' + newname + '".')


    ### Adding TreeObjs to the MultiTreeHolder


    def add_tree_obj(self, treeobj:TreeObj, nickname:str=None):
        if not self.check_nickname_is_valid(nickname):
            return
        if self.treedict.get(nickname, None):
            print('Tree with name "' + nickname + '" already in session. Use "droptree" to drop this tree or "rename" to rename it.')
            return False
        treeobj.deactivate()
        treeobj.set_offsystem()
        treeobj.session = self.session
        self.treedict[nickname] = treeobj
        if (self.CTREE==None) or (self.CTREE == self.SYSCHAR):
            self.CTREE = nickname
        return True

    def load_tree_file(self, filename:str):
        pathlike = Path(filename).resolve()
        if not pathlike.exists():
            print('Error locating file ' + str(pathlike))
            return False
        try:
            with open(filename, mode='r') as fopen:
                try:
                    file_dict = json.load(fopen)
                except:
                    print('Error opening file ' + filename)
                    return False
        except Exception as error:
            print(error)
            return False
        T = TreeObj()
        if not file_dict.get('data', False):
            if file_dict['type']==DEVINFO:
                dictin = file_dict
                settings = dict()
            else:
                print('Error loading file: cannot find DEVINFO of json file.')
                return
            print('Old format file loaded.')
        else:
            dictin = file_dict['data']
            settings = file_dict['settings']
        T.load_dict(dictin)
        T.load_settings(settings)
        T.set_savefile(filename, quiet=True, force=True)
        T.has_unsaved_changes = False  
            # we loaded from this file, so we know the info is there
        return T

    def load_tree(self, filename:str, nickname:str=None):
        T = self.load_tree_file(filename)
        if not T:
            return False
        if nickname == None:
            nickname = T.dict['device']
        self.add_tree_obj(T, nickname=nickname)
        return True

    def new_tree(self, filename:str, device:str, nickname:str=None):
        T = TreeObj(session=self.session, onsystem=False, device=device)
        T.set_savefile(filename)
        if not nickname:
            nickname = T.dict['device']
        T.has_unsaved_changes = True
        self.add_tree_obj(T, nickname=nickname)

    @checkname(allowsys=False, replace=False)
    def droptree(self, treename:str=None, force:bool=True):
        if not self.treedict[treename].has_unsaved_changes or force:
            self.treedict.pop(treename)
            print("Tree " + treename + " dropped...")
            if self.CTREE==treename:
                self.CTREE = self.SYSCHAR
            return True
        else:
            print("Error: Tree has unsaved changes.")
            print('Use the -f flag to force the drop.')
            return False


    ### Basic Unix-like commands for trees


    @checkname(allowsys=True, replace=True)
    def ls_tree(self, treename:str=None, path:str=None, recorded_only:bool=False, show_mtime:bool=False):
        if treename==self.SYSCHAR:
            ls_os(path, show_mtime)
        else:
            self.treedict[treename].ls(path, recorded_only, show_mtime)

    def ls(self, path:str=None):
        if self.CTREE==self.SYSCHAR:
            ls_os(path)
        else:
            self.treedict[self.CTREE].ls(path)

    def pwt(self):
        """Print Working Tree"""
        if self.CTREE == None:
            print('No tree is currently active.')
            return
        elif self.CTREE == self.SYSCHAR:
            print('Currently on System.')
        else:
            print('Current Tree: ' + self.CTREE)

    @checkname(allowsys=True, replace=True)
    def pwd(self, treename:str=None, full:bool=False):
        if treename==self.SYSCHAR:
            print(os.getcwd())
        else:
            self.treedict[treename].pwd(full)

    @checkname(allowsys=True, replace=True)
    def print_tree(self, treename:str=None, print_session=True):
        if treename==self.SYSCHAR:
            print('>> System')
            return
        indicator = (">> " if self.CTREE==treename else "   ")
        device = self.treedict[treename].dict.get('device', 'N/A')
        hash_file = ("Yes" if self.treedict[treename].hash_files else "No")
        session = (self.treedict[treename].session if self.treedict[treename].session else 'None')
        savefile = (self.treedict[treename].savefile if self.treedict[treename].savefile else 'None')
        print()
        print(indicator +
              'Nickname: ' + treename + '\n' +
              "   Device: " + (device if device!=None else 'N/A') + '\n' +
              "   Last Modified: " + time_formatted(self.treedict[treename].dict['mtime']) + '\n' + 
              "   Hash files?: " + hash_file + '\n' + 
              "   Save File: " + str(savefile) + '\n' +
              (('   Session: ' + session) if print_session else ''))

    def list_trees(self):
        print()
        print('Session: ' + (self.session if self.session else 'None'))
        for treename in self.treedict:
            self.print_tree(treename, print_session=False)
    
    @checkname(allowsys=False, replace=True)
    def print_savefile(self, treename:str=None):
        savefile = self.treedict[treename].savefile
        print('Savefile for "' + treename + '" :')
        print(savefile)

    @checkname(allowsys=True, replace=True)
    def cd_tree(self, treename:str=None, path:str=None):
        if not path:
            print("Must input path to change directories")
            return False
        if treename==self.SYSCHAR:
            try:
                os.chdir(path)
                return True
            except PermissionError:
                print('Permission denied.')
                return False
            except Exception as error:
                print('Error: ', error)
                return False
        else:
            return self.treedict[treename].cd(path)

    @checkname(allowsys=True, replace=False)
    def chtree(self, treename):
        self.CTREE = treename


    ### Recording entries


    @checkname(allowsys=False, replace=True)
    def recroot(self, 
            treename:str=None,
            localpath:str='.',
            nickname:str=None,
            treepath:str=None):
        self.treedict[treename].recroot(localpath, nickname, treepath)

    @checkname(allowsys=False, replace=True)
    def rec(self, treename:str=None, objpath:str='.', verbose:bool=False):
        return self.treedict[treename].rec(objpath, verbose=verbose)

    @checkname(allowsys=False, replace=True)
    def rrec(self, treename:str=None, objpath:str='.', levels:int=0, verbose_levels:int=0):
        return self.treedict[treename].rrec(objpath, levels, global_ignore=self.ignore_names_list, verbose_levels=verbose_levels)

    @checkname(allowsys=False, replace=True)
    def lrec(self, treename:str=None, objpath:str='.', verbose:bool=False):
        return self.treedict[treename].lrec(objpath, verbose=verbose)

    @checkname(allowsys=False, replace=True)
    def remove(self, treename:str=None, objpath:str='.'):
        self.treedict[treename].remove(objpath)


    ### global ignore list
        
    
    def ignorelist_print(self):
        if len(self.ignore_names_list)==0:
            print(' [ none ] ')
            return
        for idx, name in enumerate(self.ignore_names_list):
            print('[', idx, ']' + '  ' + name)
    
    def ignorelist_remove(self, name:str=None, idx:int=None):
        if name:
            for ii, entry in enumerate(self.ignore_names_list):
                if entry==name:
                    self.ignore_names_list.pop(ii)
                    return True
            print('Name "' + name + '" not found in Ignore List.')
            return False
        elif idx is not None:
            if idx >= len(self.ignore_names_list):
                print('Error: index (' + idx + ') too large for Ignore List.')
                return False
            self.ignore_names_list.pop(idx)
            return True
        else:
            print('Invalid input for ignorelist_remove')
            return False

    def ignorelist_clear(self):
        self.ignore_names_list = []
    
    def ignorelist_add(self, name:str=None):
        if not name:
            print('Ignore List name cannot be empty.')
            return False
        self.ignore_names_list.append(name)
        return True

    @checkname(allowsys=False, replace=True)
    def ignorelist_add_local(self, treename:str=None, name:str=None):
        self.treedict[treename].ignorelist_add(name)
    
    @checkname(allowsys=False, replace=True)
    def ignorelist_print_local(self, treename:str=None):
        self.treedict[treename].ignorelist_print()
    
    @checkname(allowsys=False, replace=True)
    def ignorelist_remove_local(self, treename:str=None, name:str=None, idx:int=None):
        self.treedict[treename].ignorelist_remove(name, idx)


    ### ### ###


    @checkname(allowsys=False, replace=True)
    def set_localpath(self, treename:str=None, root:str=None, path:str=None):
        if not root:
            print('Error: must give a root name.')
            return
        if not path:
            print('Error: localpath must not be empty.')
            return
        self.treedict[treename].root_set_localpath(root, path)

    @checkname(allowsys=False, replace=True)
    def set_os_type(self, treename:str=None, os_type:str=None):
        self.treedict[treename].set_os_type(os_type)
        print('Tree "' + treename + '" has os type set to "' + str(self.treedict[treename].print_os_type()) + '".')

    @checkname(allowsys=True, replace=True)
    def print_os_type(self, treename:str=None):
        if treename==self.SYSCHAR:
            print(os.name)
        else:
            print(self.treedict[treename].print_os_type())

    @checkname(allowsys=False, replace=True)
    def update(self, treename:str=None, objpath:str='.'):
        self.treedict[treename].update(objpath)

    @checkname(allowsys=False, replace=True)
    def r_update(self,
            treename:str=None,
            objpath:str='.',
            levels:int=0):
        self.treedict[treename].r_update(objpath, levels)

    @checkname(allowsys=False, replace=False)
    def tie(self, treename, Cfile, Dfile):
        """
        Ties together Cfile in the current tree's current directory
        with Dfile in the other tree's current directory.
        """
        result, Cdict, Ddict = self.tie_header(treename, Cfile, Dfile)
        if result is None:
            # treename not valid
            return
        if result==2:
            print('Error: cannot tie a file to itself.')
            return
        if result==0:
            self.tie_guts(treename, Cdict, Ddict)

    def tie_guts(self, treename, Cdict, Ddict):
        succeeded = tie_files(Cdict, Ddict, session=self.session)
        if succeeded:
            self.treedict[self.CTREE].has_unsaved_changes = True
            self.treedict[treename].has_unsaved_changes = True
    
    def tie_header(self, treename, Cfile, Dfile):
        """
        Turns the paths into dicts, for use in tie files.
        Return codes:
          0 - success
          1 - failed to find one of the objects
          2 - attempted to tie a file to itself
        """
        Cdict = self.treedict[self.CTREE].traverse_path(Cfile)
        if not Cdict:
            # traverse_path() already has some failure text,
            # this is printed in addition to it
            print('   (In tree ' + self.CTREE + ')')
            return 1, dict(), dict()
        Ddict = self.treedict[treename].traverse_path(Dfile)
        if not Ddict:
            print('   (In tree ' + treename + ')')
            return 1, dict(), dict()
        if Cdict is Ddict:
            return 2, Cdict, Ddict
        return 0, Cdict, Ddict

    @checkname(allowsys=False, replace=True)
    def a_tie(self, treename, Cdir, Ddir):
       """
       Ties together all entries in Cdir and Ddir which share a name.
       """
       result, Cdict, Ddict =  self.tie_header(treename, Cdir, Ddir)
       if result==1:
           return
       if result==2:
           print('Error: cannot tie a file to itself.')
           return
       if result==0:
           if Cdict['type'] not in DIR_LIST_TYPES or Ddict['type'] not in DIR_LIST_TYPES:
               print('Error: both entries must be directories.')
               return
           all_c_entries = [entry 
                   for entry in Cdict['dir_list']+Cdict['file_list'] 
                   if entry['type']!='unrecorded']
           all_d_entries = [entry 
                   for entry in Ddict['dir_list']+Ddict['file_list'] 
                   if entry['type']!='unrecorded']
           for c_entry in all_c_entries:
               for d_entry in all_d_entries:
                   if c_entry['name']==d_entry['name']:
                       self.tie_guts(treename, c_entry, d_entry)


    @checkname(allowsys=False, replace=True)
    def add_note(self, treename:str=None, file:str=None, note_text:str=None):
        if not file:
            print("Error: file name required.")
            return
        if not note_text:
            print("Error: note text cannot be empty.")
            return
        self.treedict[treename].add_note(file, note_text)

    @checkname(allowsys=False, replace=True)
    def remove_note(self, treename:str=None, file:str=None, note_idx:int=None):
        if not file:
            print("Error: file name required.")
            return
        if note_idx is None:
            print("Error: note index required.")
            return
        self.treedict[treename].remove_note(file, note_idx)

    @checkname(allowsys=True, replace=True)
    def info(self, treename:str=None, filename:str='.', option:str=None, sys_do_hash:bool=False):
        if treename==self.SYSCHAR:
            print_os_file_info(filename, do_hash=sys_do_hash)
        else:
            self.treedict[treename].info(filename, option=option)
    
    @checkname(allowsys=False, replace=True)
    def history(self, treename:str=None, filename:str='.', full:bool=False):
        self.treedict[treename].history(filename, full)

    @checkname(allowsys=False, replace=True)
    def notes(self, treename:str=None, filename:str='.'):
        self.treedict[treename].notes(filename)

    @checkname(allowsys=False, replace=True)
    def ties(self, treename:str=None, filename:str='.'):
        self.treedict[treename].ties(filename)





# hist_events.py
# Deifnitions of history events, for appending to files' history.


import time
from treerec.trace_ancestors import *
from treerec.pathtuputils import subtract_paths


def record_entry(dictin:dict, session:str=None, localpath:str=None):
    event = {
        'type': 'record',
        'name': dictin['name'],
        'device': dictin['device'],
        'hash': dictin['hash']
        }
    if localpath:
        event['localpath'] = localpath
    return {
        'time' : int(time.time()),
        'session' : session,
        'name' : dictin['name'],
        'event' : event
    }


def move_entry(source, destination, session=None):
    # remember to fix broken ties, or leave a log for the inactive trees to fix it
    pass


def check_tied(source:dict, destination:dict):
    # Check the ties of two entries to see if they are tied.
    # Returns True if *both files* have ties pointing to each other,
    #   and False otherwise.
    D_uuid = get_tree_id(destination)
    S_uuid = get_tree_id(source)
    S_in_D = False
    D_in_S = False
    for tie_entry in destination['ties']:
        if tie_entry['uuid']==S_uuid and tie_entry['path']==source['path']:
            S_in_D = True
            break
    for tie_entry in source['ties']:
        if tie_entry['uuid']==D_uuid and tie_entry['path']==destination['path']:
            D_in_S = True
            break
    if S_in_D and D_in_S:
        return True
    else:
        return False


def compare_hashes(dictA:dict, dictB:dict):
    """
    Return True if the hashes are identical, and not a fail value
    (e.g. hash == 0, -1).
    """
    if type(dictA.get('hash', 0)) is str and type(dictB.get('hash', 0)) is str:
        return dictA['hash']==dictB['hash']
    return False


def tie_entry(source:dict, destination:dict, session:str=None):
    # First, check to see if the files are already tied
    if check_tied(source, destination):
        return False
    # Else, return the tie event
    root = find_current_root(destination)
    after_root_path = subtract_paths(root['path'], destination['path'])
    root_nickname = root.get('nickname', root['name'])
    nickname_path = (root_nickname, *after_root_path)
    return {
        'time' : int(time.time()),
        'session' : session,
        'name' : source['name'],
        'event' : {
            'type' : 'tie',
            'source' : source['path'],
            'name' : destination['name'],
            'device' : destination['device'],
            'uuid' : get_tree_id(destination),
            'session' : (session if session else destination.get('session', None)),
            'time': int(time.time()),
            'path' : destination['path'],
            'nickname_path': nickname_path,
            'hash': destination.get('hash', 0),
            'identical_hash': compare_hashes(source, destination)
        }
    }


def update_entry(dictin:dict, update_info:dict, session:str=None):
    update_fields = [attr for attr in update_info]
    return {
        'time' : int(time.time()),
        'session' : session,
        'name' : dictin['name'],
        'event' : {
            'type' : 'update',
            'old' : {key: dictin[key] for key in update_fields},
            'new' : {key: update_info[key] for key in update_fields}
            }
        }



import win32security
def get_win_info(entry):
    # Thanks to Wendel at https://stackoverflow.com/questions/71018167/get-owner-of-a-file-using-python-on-windows-system
    sd = win32security.GetFileSecurity(str(entry), win32security.OWNER_SECURITY_INFORMATION)
    owner_sid = sd.GetSecurityDescriptorOwner()
    name, domain, tpe = win32security.LookupAccountSid(None, owner_sid)
    return name, domain



# entryutils.py
# Utilities for entry-level operations
# Includes: tying files, updating files, printing file info.


from treerec.primitives import *
from treerec.hist_events import *
from treerec.dictutils import *
from treerec.dirutils import *
from treerec.pathtuputils import *
from treerec.navutils import *
import time, datetime


def std_print(entry:dict, field:str, failstr:str=FIELD_NOT_RECORDED):
    # The point of this is that it prints 'N/A' if *either* the field
    # is not present, or if it's present with a value of None or False.
    return (entry[field] if entry.get(field, False) else failstr)


def time_formatted(itime):
    if itime==-1:
        return ''
    return datetime.datetime.fromtimestamp(int(itime)).strftime(DATE_FORMAT + '  ' + TIME_FORMAT)


def std_time(entry, time_key:str, failstr:str=FIELD_NOT_RECORDED):
    return (time_formatted(entry[time_key]) if entry.get(time_key, False)
            else failstr)

def std_hash(entry:dict, failstr:str=FIELD_NOT_RECORDED, abbrev:bool=False):
    hash_data = entry.get('hash', False)
    if hash_data==-1:
        return NO_HASH_BC_DIR
    elif hash_data==-2:
        return NO_HASH_BC_PERMISSIONS
    elif not hash_data:
        return failstr
    elif abbrev:
        if len(hash_data)>=7:
            return hash_data[0:7]
        else:
            return hash_data
    else:
        return hash_data


def std_hist_session(entry, idx:int, failstr:str=FIELD_NOT_RECORDED):
    try:
        session = (entry['history'][idx]['session'] if entry['history'][idx].get('session', False)
                else failstr)
        mtime = (entry['history'][idx]['time'] if entry['history'][idx].get('time', False) else '')
        return '"' + session + '" @ ' + time_formatted(mtime)
    except:
        return failstr


def print_dict_info(
        entry,
        data=('owner', 'device',
              'nicknamepath', 'treepath', 'os_type', 'ctime', 'mtime', 'atime','hash',
              'notes', 'ties', 'history', 'recordedsession', 'lastmodifiedsession')):
    print("  Name: " + entry['name'])
    print("  Type: " + entry['type'])
    if entry['type'] == 'unrecorded':
        return
    SP = SuperPath(entry)
    printouts = {
        'owner': "  Owner: " + std_print(entry, 'owner'),
        'device': "  Device: " + std_print(entry, 'device'),
        'nicknamepath': "- Path: (" + SP.root_nickname + ")/" + SP.print_after_root(),
        'treepath': "- Tree Path: " + SP.print_path(),
        'os_type': "- OS Type: " + std_print(entry, 'os_type'),
        'ctime': "- C-time: " + std_time(entry, 'ctime'),
        'mtime': "- Modified: " + std_time(entry, 'mtime'),
        'atime': "- Accessed: " + std_time(entry, 'atime'),
        'hash': '- Hash: ' + std_hash(entry),
        'abbrev-hash': '- Hash: ' + std_hash(entry, abbrev=True),
        'notes': "> Notes: " + str(len(entry.get('notes', []))),
        'ties': "> Ties: " + str(len(entry.get('ties', []))),
        'history': "> History: " + str(len(entry.get('history', []))),
        'recordedsession': '> Recorded: Session ' + std_hist_session(entry, 0),
        'lastmodifiedsession': '> Last modified: Session ' + std_hist_session(entry, -1)
    }
    for key in data:
        print(printouts[key])

def print_os_file_info(filename,
                       do_hash=True,
                       data=('name', 'owner', 'type', 'path', 'os_type',
                             'ctime', 'mtime', 'atime', 'hash')):
    pathlike = Path(filename).resolve().absolute()
    if not pathlike.exists():
        print('Error locating file ' + str(pathlike))
        return
    stat = pathlike.stat()
    owner = fileobj_get_owner(pathlike)
    hash_data = (hash_file(pathlike) if do_hash else 0)
    if hash_data==-1:
        hash_print = NO_HASH_BC_DIR
    elif hash_data==0:
        hash_print = FIELD_NOT_RECORDED
    else:
        hash_print = hash_data
    printouts = {
        'name': '  Name: ' + pathlike.name,
        'owner': '  Owner: ' + owner,
        'type': '  Type: ' + ('dir' if pathlike.is_dir() else 'file'),
        'path': '  Path: ' + str(pathlike),
        'os_type': '  OS Type: ' + os.name,
        'ctime': '- C-time: ' + time_formatted(stat.st_ctime),
        'mtime': '- Modified: ' + time_formatted(stat.st_mtime),
        'atime': '- Accessed: ' + time_formatted(stat.st_atime),
        'hash': '- Hash: ' + hash_print 
    }
    for key in data:
        print(printouts[key])


def new_entry_in_dir(parent_dict, objname, device=None, os_type=None, session=None, quiet=False, do_hash=True):
    """
    Records a new entry in the given directory.
    Converts treepath into localpath, pulls system information, then records treepath to new entry.
    """
    parent_localpath = combine_pathtuples(get_localpath_from_path(parent_dict))
    obj_pathlike = Path(parent_localpath).resolve().absolute() / objname
    new_dict = create_dict(obj_pathlike, 
                           rootname=parent_dict.get('root'), 
                           session=session, 
                           device=device, 
                           os_type=os_type,
                           do_hash=do_hash)
    if new_dict==False:
        return False
    newdict_treepath = (*parent_dict['path'], objname)
    new_dict['path'] = newdict_treepath
    return replace_entry_in_dir(parent_dict, new_dict, quiet=quiet)


def update_unrecorded_contents(dictin:dict, pathlike:Path=None, quiet:bool=False):
    """
    Updates the contents of a directory dict with the contents of the directory at pathlike.
    Adds new unrecorded files, and removes unrecorded files present in tree but not in pathlike.
    """
    if not pathlike:
        pathlike = Path(combine_pathtuples(calculate_localpath(dictin)))
    if dictin['type']==DEVINFO:
        # DEVINFO has children which are not in a tree structure mirrored
        # by the tree structure on the system
        # I.e., the roots are in locations all over the place.
        return False
    if dictin['type'] not in DIR_LIST_TYPES:
        if not quiet: print("Cannot update contents of non-directory.")
        return False
    try:
        stat = pathlike.stat()
    except FileNotFoundError:
        if not quiet: print("File not found at path " + str(pathlike))
        return False
    except PermissionError:
        print('Permission denied.')
        return False
    except:
        print('Unable to update unrecorded contents.')
        return False
    contents_list = [pathlike/entry
            for entry in os.listdir(pathlike)]
    dict_dir_names = [entry['name'] 
            for entry in dictin['dir_list']]
    dict_file_names = [entry['name'] 
            for entry in dictin['file_list']]
    sys_dir_names = [entry.name
            for entry in contents_list 
            if entry.is_dir()]
    sys_file_names = [entry.name 
            for entry in contents_list 
            if not entry.is_dir()]
    # check for missing dirs in dict
    for entry in sys_dir_names:
        if entry not in dict_dir_names:
            dictin['dir_list'].append( {
                    'name': entry, 
                    'type': 'unrecorded'
                    } )
    # check for missing dirs in sys
    for idx in reversed(range(len(dictin['dir_list']))):
        entry = dictin['dir_list'][idx]
        if entry['type']=='unrecorded' and entry['name'] not in sys_dir_names:
            dictin['dir_list'].pop(idx)
    # check for missing files in dict
    for entry in sys_file_names:
        if entry not in dict_file_names:
            dictin['file_list'].append( {
                'name': entry,
                'type': 'unrecorded'
                } )
    # check for missing files in sys
    for idx in reversed(range(len(dictin['file_list']))):
        entry = dictin['file_list'][idx]
        if entry['type']=='unrecorded' and entry['name'] not in sys_file_names:
            dictin['file_list'].pop(idx)
    # alphebatize the results:
    dictin['dir_list'] = sorted(dictin['dir_list'], key=lambda x: str.casefold(x['name']))
    dictin['file_list'] = sorted(dictin['file_list'], key=lambda x: str.casefold(x['name']))
    """
    for entry in contents_list:
        if entry.is_dir():
            if entry.name not in dict_dir_names:
                dictin['dir_list'].append({'name': entry.name, 'type': 'unrecorded'})
        elif entry.name not in dict_file_names:
            dictin['file_list'].append({'name': entry.name, 'type': 'unrecorded'})
    # if we find an unrecorded entry in dictin which is not in contents_list, remove it:
    for x_list in ['dir_list', 'file_list']:
        for idx in reversed(range(len(dictin[x_list]))):
            entry = dictin[x_list][idx]
            if entry['type']=='unrecorded' and (entry['name'] not in [x.name for x in contents_list]):
                dictin[x_list].pop(idx)
    """
    return dictin

def update_object(dictin:dict, pathlike:Path=None, session:str=None, do_hash:bool=True, quiet:bool=False):
    if dictin['type']=='unrecorded':
        return False, False
    if dictin['type']==DEVINFO:
        # DEVINFO is not a directory on the system.
        if not quiet: print('Cannot update DEVINFO: it is not a system directory.')
        return False, False
    if not pathlike:
        pathlike = Path(combine_pathtuples(calculate_localpath(dictin)))
    try:
        stat = pathlike.stat()
    except FileNotFoundError:
        if not quiet: print("File not found at path " + str(pathlike))
        return False, False
    update_info = {
        'name' : dictin['name'],
        'owner' : fileobj_get_owner(pathlike),
        'ctime' : int(stat.st_ctime),
        'mtime' : int(stat.st_mtime),
        'atime' : int(stat.st_atime),
        'hash' : (hash_file(pathlike) if do_hash else 0)
    }
    identical = True
    for attr in update_info:
        if dictin.get(attr, None) != update_info[attr]:
            identical = False
            break
    if identical:
        if not quiet: print('%%% is identical.')
        return dictin, False
    dictin['history'].append(update_entry(dictin, update_info, session=session))
    for attr in update_info:
        dictin[attr] = update_info[attr]
    if dictin['type'] in DIR_LIST_TYPES:
        dictin = update_unrecorded_contents(dictin, pathlike, quiet)
    return dictin, True


def tie_files(dict1, dict2, session=None):
    if dict1['type']=='unrecorded':
        print("Cannot tie unrecorded file: " + dict1['name'])
        return False
    if dict2['type']=='unrecorded':
        print("Cannot tie unrecorded file: " + dict2['name'])
        return False

    dict1_event = tie_entry(dict1, dict2, session=session)
    if not dict1_event:
        print('Files are already tied.')
        return False
    dict2_event = tie_entry(dict2, dict1, session=session)
    dict1_tie = dict1_event['event']
    dict2_tie = dict2_event['event']
    print('.. ' 
            + dict2_tie['nickname_path'][0] + '|'
            + dict2_tie['name'] + ' -> '
            + dict1_tie['nickname_path'][0] + '|'
            + dict1_tie['name']
            )

    dict1['ties'].append(dict1_tie)
    dict2['ties'].append(dict2_tie)
    dict1['history'].append(dict1_event)
    dict2['history'].append(dict2_event)

    return True


def print_update_event_file_data(sub_event:dict, diff_list:list=None):
    def indent(key):
        if diff_list:
            charout = '   ' + ('*' if key in diff_list else ' ')
        else:
            charout = '    '
        return charout

    return (
            indent('owner') + "Owner: " + std_print(sub_event, 'owner') + '\n' +
            indent('ctime') + "C-time: " + std_time(sub_event, 'ctime') + '\n' +
            indent('mtime') + "Modified: " + std_time(sub_event,'mtime') + '\n' +
            indent('atime') + "Accessed: " + std_time(sub_event,'atime') + '\n' +
            indent('hash') + "Hash: " + std_hash(sub_event) + '\n'
        )


def print_update_event_data_with_diff(event:dict):
    """
    Prints all of the fields for an update event,
    with an '*' next to those fields which changed.
    """
    diff_list = [key for key in event['old'] if event['old'][key]!=event['new'].get(key,None)]
    return (
        '--Old: ' + '\n' + print_update_event_file_data(event['old'], diff_list=None) +
        '--New: ' + '\n' + print_update_event_file_data(event['new'], diff_list) + '\n')


def print_history_event(dictin:dict, idx:int):
    if idx>=len(dictin['history']):
        return False
    histevent = dictin['history'][idx]
    event = histevent['event']
    header = (">> " + std_time(histevent, 'time') + ', ' +
              'Session: ' + std_print(histevent, 'session') + '\n')
    if event['type']=='record':
        return header + ('Recorded ' + std_print(event, 'name') + '\n' +
                         'Device: ' + std_print(event, 'device') + '\n' +
                         'Local path: ' + std_print(event, 'localpath') + '\n' +
                         'Hash: ' + std_hash(event) + '\n')
    elif event['type']=='tie':
        return header + ('Tied \n' +
                         std_print(dictin, 'device') + ': ' + combine_pathtuples(event['source']) + '\n' +
                         'to \n' +
                         std_print(event, 'device') + ': ' + combine_pathtuples(event['path']) + '\n')
    elif event['type']=='update':
        return header + ('Updated ' + std_print(histevent, 'name') + '\n' + 
                         print_update_event_data_with_diff(event))


def print_dict_history_full(dictin:dict):
    for idx, _ in enumerate(dictin['history']):
        print(print_history_event(dictin, idx))



from treerec.command_parser import *
from copy import copy


helpflags = ('-h', '--help', '--h', '-help')


cmd_blank = 'blank'
cmd_quit = 'quit'
cmd_help = 'help'
cmd_addtree = 'addtree'
cmd_create_tree = 'createtree'
cmd_set_session_name = 'sessionname'
cmd_activate_tree = 'activate_tree'
cmd_deactivate_tree = 'deactivate_tree'
cmd_echo = 'echo'
cmd_list_all_trees = 'list_trees'
cmd_list_tree = 'list_tree'
cmd_on_system = 'on_system'
cmd_set_savefile = 'set_savefile'
cmd_ls = 'ls'
cmd_cd = 'cd'
cmd_cdls = 'cdls'
cmd_recroot = 'recroot'
cmd_set_localpath = 'root_localpath'
cmd_rec = 'rec'
cmd_remove = 'remove'
cmd_save = 'save'
cmd_rename = 'rename'
cmd_droptree = 'droptree'
cmd_pwt = 'pwt'
cmd_change_tree = 'chtree'
cmd_tie = 'tie'
cmd_info = 'info'
cmd_history = 'history'
cmd_print_ties = 'ties'
cmd_set_hash = 'sethash'
cmd_set_os_type = 'set_os_type'
cmd_pwd = 'pwd'
cmd_update = 'update'
cmd_notes = 'note'
cmd_remove_note = 'remove_note'
cmd_autoupdate = 'auto_update'
cmd_ignore = 'ignore_list'
## Include rec -a flags for recording only files / directories
##      -> update command_parser to include an InputParam that takes an
##          arbitrary number of inputs
## Include a ls flag (-o <var>) for sorting outputs


BLANK = CommandInput(
    '', ' ', '\t', '\n', '\r', '\r\n', '\n\r', '\r\r', '\n\n',
    default_data=blank_dict(cmd_blank),
    final=True,
    terminal=True
)


QUIT = CommandInput(
    'quit', 'exit', 'q', 'bye', 'goodbye',
    include_help=True,
    final=True,
    terminal=True,
    action_func=mykey_func('command', cmd_quit),
)


ECHO = CommandInput(
    'echo',
    include_help=True,
    default_data={
        'command': cmd_echo,
        'text': None,
    },
    children=[
        OptionalVariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('text')
        )
    ]
)


CMD_HELP = CommandInput(
    'help', 'h',
    include_help=False,
    default_data={
        'command': cmd_help,
        'help': True
    },
    terminal=True
)


SET_SESSION_NAME = CommandInput(
    'sessionname', 'session',
    include_help=True,
    default_data=blank_dict(cmd_set_session_name, 'text'),
    children=[
        VariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('text')
        )
    ]
)

LIST_ALL_TREES = CommandInput(
    'lstrees', 'trees', 'list',
    include_help=True,
    default_data=blank_dict(cmd_list_all_trees),
    terminal=True  # command_parser is set to still accept flags, even if a command is terminal
)


LIST_TREE = CommandInput(
    'lstree', 'tree',
    include_help=True,
    default_data=blank_dict(cmd_list_tree, 'text'),
    children=[
        OptionalVariableInputParam(
            terminal=True,
            action_func=theirkey_func('text')
        )
    ]
)


ADD_TREE = CommandInput(
    'addtree', 'add_tree', 'add', 'load',
    include_help=True,
    default_data={
        'command': cmd_addtree,
        'filename': None,
        'nickname': None,
        'newtree': False,
        'device': None
    },
    flags=[
        InputParam(
            '-c', '--new', '--create',
            final=True,
            action_func=mykey_func('newtree', True)
        ),
        InputParam(
            '-d', '--dev', '-dev', '--device',
            final=False,
            action_func=blank_func(),
            children=[
                VariableInputParam(
                    final=True,
                    terminal=False,
                    action_func=theirkey_func('device')
                )
            ]
        ),
        InputParam(
            '-n', '--name', '--nickname',
            final=False,
            action_func=blank_func(),
            children=[
                VariableInputParam(
                    final=True,
                    terminal=False,
                    action_func=theirkey_func('nickname')
                )
            ]
        )
    ],
    children=[
        VariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('filename'),
        )
    ]
)


CREATE_TREE = CommandInput(
        'create', 'new',
        include_help=True,
        default_data={
            'command': cmd_create_tree,
            'nickname': None,
            'device': None,
            'filename': None
            },
        children=[
            VariableInputParam(
                action_func=theirkey_func('nickname'),
                children=[
                    VariableInputParam(
                        action_func=theirkey_func('device'),
                        children=[
                            VariableInputParam(
                                action_func=theirkey_func('filename'),
                                terminal=True
                                )
                            ]
                        )
                    ]
                )
            ]
        )


RENAME_TREE = CommandInput(
    'rename', 'name',
    include_help=True,
    default_data = {
        'command': cmd_rename,
        'name1': None,
        'name2': None
    },
    children = [
        VariableInputParam(
            action_func=theirkey_func('name1'),
            children=[
                OptionalVariableInputParam(
                    terminal=True,
                    action_func=theirkey_func('name2')
                )
            ]
        )
    ]
)


ACTIVATE_TREE = CommandInput(
    'activate', 'active',
    include_help=True,
    default_data={
        'command': cmd_activate_tree,
        'name': None,
        'session': None
    },
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('name'),
            children=[
                OptionalVariableInputParam(
                    final=True,
                    terminal=True,
                    action_func=theirkey_func('session')
                )
            ]
        )
    ]
)


DEACTIVATE_TREE = CommandInput(
    'deactivate', 'inactive',
    include_help=True,
    default_data={
        'command': cmd_deactivate_tree,
        'name': None
    },
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('name'),
            final=True,
            terminal=True
        )
    ]
)


ON_SYSTEM = CommandInput(
    'onsystem', 'on_system', 'on',
    include_help=True,
    default_data={'command': cmd_on_system,
                  'name': None,
                  'action': 'on',
                  'device': None},
    flags=[
        InputParam(
            '-d', '--device',
            action_func=blank_func(),
            children=[
                VariableInputParam(
                    final=True,
                    terminal=False,
                    action_func=theirkey_func('device')
                )
            ]
        )
    ],
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('name'),
            children=[
                OptionalVariableInputParam(
                    final=True,
                    terminal=True,
                    action_func=theirkey_func('action')
                )
            ]
        )
    ]
)

OFF_SYSTEM = CommandInput(
    'offsystem', 'off_system', 'off',
    include_help=True,
    default_data={'command': cmd_on_system, 'name': None, 'action': 'off'},
    children=[
        OptionalVariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('name')
        )
    ]
)

SET_SAVE_FILE = CommandInput(
    'setsave', 'savefile', 'set_savefile',
    include_help=True,
    default_data={
        'command': cmd_set_savefile,
        'name1': None,
        'name2': None,
        'force': False,
        'show': False
    },
    flags=[
        CommandInput(
            '-f', '--force',
            action_func=mykey_func('force', True),
            final=True
        ),
        CommandInput(
            '-s', '--show',
            action_func=mykey_func('show', True),
            final=False,
            children=[
                OptionalVariableInputParam(
                    action_func=theirkey_func('name1'),
                    final=True,
                    terminal=True
                )
            ]
        )
    ],
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('name1'),
            children=[
                OptionalVariableInputParam(
                    final=True,
                    terminal=True,
                    action_func=theirkey_func('name2')
                )
            ]
        )
    ]
)


SET_OS_TYPE = CommandInput(
        'os', 'set_os', 'setos',
        include_help=True,
        default_data={
            'command': cmd_set_os_type,
            'input1': None, 
            'input2': None,
            'reset': False,
            'show': False
            },
        flags=[
            CommandInput(
                '-r',
                action_func=mykey_func('reset', True),
                children=[
                    OptionalVariableInputParam(
                        action_func=theirkey_func('input1'),
                        final=True,
                        terminal=True
                        )
                    ]
                ),
            CommandInput(
                '-s',
                action_func=mykey_func('show', True),
                children=[
                    OptionalVariableInputParam(
                        action_func=theirkey_func('input1'),
                        final=True,
                        terminal=True
                        )
                    ]
                )
            ],
        children=[
            OptionalVariableInputParam(
                action_func=theirkey_func('input1'),
                children=[
                    OptionalVariableInputParam(
                        action_func=theirkey_func('input2'),
                        final=True,
                        terminal=True
                        )
                    ]
                )
            ]
        )


SET_HASH = CommandInput(
    'sethash', 'hash',
    include_help=True,
    default_data={
        'command': cmd_set_hash,
        'action': None
    },
    children=[
        VariableInputParam(
            action_func=theirkey_func('action'),
            final=True,
            terminal=True
        )
    ]
)


CMD_CD = CommandInput(
    'cd',
    include_help=True,
    default_data={
        'command': cmd_cd,
        'first': None,
        'second': None
    },
    children=[
        VariableInputParam(
            action_func=theirkey_func('first'),
            children=[
                OptionalVariableInputParam(
                    final=True,
                    terminal=True,
                    action_func=theirkey_func('second')
                )
            ]
        )
    ]
)

CMD_LS = CommandInput(
    'ls', 'dir',
    include_help=True,
    default_data={
        'command': cmd_ls,
        'recorded_only': False,
        'mtime': False,
        'first': None,
        'second': None
    },
    flags=[
        InputParam(
            '-r', '--recorded',
            action_func=mykey_func('recorded_only', True),
            final=True
        ),
        InputParam(
            '-m',
            action_func=mykey_func('mtime', True),
            final=True
        )
    ],
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('first'),
            children=[
                OptionalVariableInputParam(
                    final=True,
                    terminal=True,
                    action_func=theirkey_func('second')
                )
            ]
        )
    ]
)


CMD_CDLS = CommandInput(
    'cdls',
    include_help=True,
    default_data = {
        'command': cmd_cdls,
        'input1': None,
        'input2': None
    },
    children=[
        VariableInputParam(
            action_func=theirkey_func('input1'),
            children=[
                OptionalVariableInputParam(
                    action_func=theirkey_func('input2'),
                    terminal=True
                )
            ]
        )
    ]
)


CMD_RECROOT = CommandInput(
    'recroot', 'newroot',
    include_help=True,
    default_data={
        'command': cmd_recroot,
        'localpath': None,
        'treepath': None,
        'name': None
    },
    flags=[
        InputParam(
            '-n', '--name', '--nickname',
            action_func=blank_func(),
            children=[
                VariableInputParam(
                    final=True,
                    action_func=theirkey_func('name')
                )
            ]
        ),
        InputParam(
            '-p', '--path', '--treepath',
            action_func=blank_func(),
            children=[
                VariableInputParam(
                    final=True,
                    action_func=theirkey_func('treepath')
                )
            ]
        )
    ],
    children=[
        VariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('localpath')
        )
    ]
)


CMD_SET_LOCALPATH = CommandInput(
    'local', 'localpath', 'set_localpath', 'root_localpath',
    include_help=True,
    default_data={
        'command': cmd_set_localpath,
        'root': None,
        'path': None
    },
    children=[
        VariableInputParam(
            action_func=theirkey_func('root'),
            children=[
                VariableInputParam(
                    action_func=theirkey_func('path'),
                    terminal=True
                )
            ]
        )
    ]
)


CMD_REC = CommandInput(
    'rec', 'record',
    include_help=True,
    default_data={
        'command': cmd_rec,
        'cd': False,
        'cdls': False,
        'path': None,
        'use_localpath': False,
        'recursive': False,
        'r_levels': 0,
        'verbose': 0,
    },
    flags=[
        InputParam(
            '-l', '--local', '--use-local', '--localpath', '--use-localpath',
            final=True,
            action_func=mykey_func('use_localpath', True)
        ),
        InputParam(
            '-r', '--recursive',
            action_func=mykey_func('recursive', True),
            children=[
                VariableInputParam(
                    final=True,
                    action_func=theirkey_func('r_levels')
                )
            ]
        ),
        InputParam(
            '-RR',
            terminal=False,
            final=True,
            action_func=(lambda x, y: x | {'recursive': True, 'r_levels': -1})
        ),
        InputParam(
            '-a', '--all',
            final=True,
            terminal=True,
            action_func=lambda params, data: params | {'recursive': True, 'r_levels': 1, 'path': '.'} 
        ),
        InputParam(
            '-cd',
            final=True,
            action_func=mykey_func('cd', True)
        ),
        InputParam(
            '-cdls',
            final=True,
            action_func=mykey_func('cdls', True)
        ),
        InputParam(
            '-v', '--verbose',
            final=True,
            action_func=mykey_func('verbose', -1)
        ),
        InputParam(
            '-vv', '--verbose-levels',
            children=[
                VariableInputParam(
                    action_func=theirkey_func('verbose'),
                    final=True
                )
            ]
        )
    ],
    children=[
        VariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('path')
        )
    ]
)


CMD_REMOVE = CommandInput(
    'remove',
    include_help=True,
    default_data={
        'command': cmd_remove,
        'confirm': False,
        'path': None
    },
    flags=[
        InputParam(
            '-y', '--yes', '--confirm',
            action_func=mykey_func('confirm', True),
            final=True
        )
    ],
    children=[
        VariableInputParam(
            action_func=theirkey_func('path'),
            terminal=True
        )
    ]
)


CMD_SAVE = CommandInput(
        'save',
        include_help=True,
        default_data={
            'command': cmd_save,
            'tree': None,
            'all': False,
            'force': False,
            'pretty': False
            },
        flags=[
            InputParam(
                '-a', '--all',
                final=True,
                action_func=mykey_func('all', True)
                ),
            InputParam(
                '-f', '--force',
                final=True,
               action_func=mykey_func('force', True)
               ),
            InputParam(
                '-p', '--pretty',
                final=True,
                action_func=mykey_func('pretty', True)
                )
            ],
        children=[
            OptionalVariableInputParam(
                final=True,
                terminal=True,
                action_func=theirkey_func('tree')
                )
            ]
        )
        

DROP_TREE = CommandInput(
    'drop', 'droptree',
    include_help=True,
    default_data={
        'command': cmd_droptree,
        'tree': None,
        'force': False
    },
    flags=[
        CommandInput(
            '-f', '--force',
            final=True,
            action_func=mykey_func('force', True)
        )
    ],
    children=[
        VariableInputParam(
            final=True,
            terminal=True,
            action_func=theirkey_func('tree')
        )
    ]
)


PRINT_WORKING_TREE = CommandInput(
    'pwt',
    include_help=True,
    default_data = {
        'command': cmd_pwt
    },
    final=True
)


PRINT_WORKING_DIRECTORY = CommandInput(
    'pwd',
    include_help=True,
    default_data={
        'command': cmd_pwd,
        'tree': None,
        'full': False,
        'local': False
        },
    flags=[
        InputParam(
            '-f', '--full',
            action_func=mykey_func('full', True),
            final=True
            ),
        InputParam(
            '-l', '--local',
            action_func=mykey_func('local', True),
            final=True
            )
        ],
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('tree'),
            final=True,
            terminal=True
            )
        ]
)

CHANGE_TREE = CommandInput(
    'chtree', 'cht', 'cdtree', 'cdt',
    include_help=True,
    default_data={
        'command': cmd_change_tree,
        'tree': None
    },
    children=[
        VariableInputParam(
            action_func=theirkey_func('tree'),
            final=True,
            terminal=True
        )
    ]
)


TIE_FILES = CommandInput(
    'tie',
    include_help=True,
    default_data={
        'command': cmd_tie,
        'all': False,
        'tree': None,
        'source': None,
        'destination': None
    },
    flags=[
        InputParam(
            '-a', '--all',
            action_func=mykey_func('all', True),
            final=True
            )
        ],
    children=[
        VariableInputParam(
            action_func=theirkey_func('source'),
            children=[
                VariableInputParam(
                    action_func=theirkey_func('tree'),
                    children=[
                        VariableInputParam(
                            action_func=theirkey_func('destination'),
                            final=True,
                            terminal=True
                        )
                    ]
                )
            ]
        )
    ]
)


CMD_INFO = CommandInput(
    'info',
    include_help=True,
    default_data={
        'command': cmd_info,
        'input1': None,
        'input2': None,
        'amount': 'short',
        'sys_hash': False,
        'history': False,
        'ties': False
    },
    flags=[
        CommandInput(
            # This is actually redundant, but I've already written it...
            '-l', '--long',
            action_func=mykey_func('amount', 'full'),
            final=True
        ),
        CommandInput(
            '-s', '--short',
            action_func=mykey_func('amount', 'short'),
            final=True
            ),
        CommandInput(
            '-H', '--do-hash',
            action_func=mykey_func('sys_hash', True),
            final=True
        ),
        CommandInput(
            '--history',
            action_func=mykey_func('history', True),
            final=True
            ),
        CommandInput(
            '--ties',
            action_func=mykey_func('ties', True),
            final=True
            )
    ],
    children=[
        VariableInputParam(
            action_func=theirkey_func('input1'),
            children=[
                OptionalVariableInputParam(
                    action_func=theirkey_func('input2'),
                    final=True,
                    terminal=True
                )
            ]
        )
    ]
)


CMD_HISTORY = CommandInput(
        'history', 'hist',
        include_help=True,
        default_data={
            'command': cmd_history,
            'full': False,
            'input1': None,
            'input2': None
            },
        flags=[
            CommandInput(
                '-f', '--full',
                action_func=mykey_func('full', True),
                final=True
                )
            ],
        children=[
            VariableInputParam(
                action_func=theirkey_func('input1'),
                children=[
                    OptionalVariableInputParam(
                        action_func=theirkey_func('input2'),
                        terminal=True
                        )
                    ]
                )
            ]
        )


CMD_TIES = CommandInput(
    'ties',
    include_help=True,
    default_data={
        'command': cmd_print_ties,
        'input1': None,
        'input2': None
        },
    children=[
        VariableInputParam(
            action_func=theirkey_func('input1'),
            children=[
                OptionalVariableInputParam(
                    action_func=theirkey_func('input2'),
                    terminal=True
                    )
                ]
            )
        ]
    )


UPDATE_ENTRY = CommandInput(
        'update',
        include_help=True,
        default_data={
            'command': cmd_update,
            'path': '.',
            'recursive': False,
            'r_levels': 0,
            'all': False
            },
        flags=[
            InputParam(
                '-r', '--recursive',
                action_func=mykey_func('recursive', True),
                children=[
                    VariableInputParam(
                        action_func=theirkey_func('r_levels'),
                        final=True
                        )
                    ]
                ),
            InputParam(
                '-a', '--all',
                action_func=mykey_func('recursive', True),
                terminal=True
                ),
            InputParam(
                '-RR',
                action_func=lambda data, y: data | {'recursive': True, 'r_levels': -1},
                terminal=True
                )
            ],
        children=[
            OptionalVariableInputParam(
                action_func=theirkey_func('path'),
                terminal=True
                )
            ]
        )


CMD_AUTOUPDATE = CommandInput(
        'autoupdate', 'auto_update',
        include_help=True,
        default_data={
            'command': cmd_autoupdate,
            'show': False,
            'input1': None,
            'input2': None,
            },
        flags=[
            InputParam(
                '-s', '--show',
                action_func=theirkey_func('show'),
                children=[
                    OptionalVariableInputParam(
                        action_func=theirkey_func('input1'),
                        terminal=True
                        )
                    ]
                )
            ],
        children=[
            OptionalVariableInputParam(
                action_func=theirkey_func('input1'),
                children=[
                    OptionalVariableInputParam(
                        'input2',
                        action_func=theirkey_func('input2'),
                        terminal=True
                        )
                    ]
                )
            ]
        )


CMD_NOTES = CommandInput(
        'note',
        include_help=True,
        default_data={
            'command': cmd_notes,
            'input1': None,
            'input2': None,
            'action': 'show'
            },
        flags=[
            CommandInput(
                '-a', '--add',
                action_func=mykey_func('action', 'add'),
                final=True
            ),
            CommandInput(
                '-r', '--remove',
                action_func=mykey_func('action', 'remove'),
                final=True
                )
            ],
        children=[
            VariableInputParam(
                action_func=theirkey_func('input1'),
                children=[
                    OptionalVariableInputParam(
                        action_func=theirkey_func('input2'),
                        terminal=True
                        )
                    ]
                )
            ]
        )


IGNORE_LIST = CommandInput(
    'ignore', 'ignorelist',
    default_data={
        'command': cmd_ignore,
        'remove': False,
        'index': False,
        'input': None
    },
    flags=[
        CommandInput(
            '-r', '--remove',
            action_func=mykey_func('remove', True),
            final=True
        ),
        CommandInput(
            '-i', '--index',
            action_func=mykey_func('index', True),
            final=True
        ),
        CommandInput(
            '-c', '--clear',
            action_func=mykey_func('clear', True),
            terminal=True
        )
    ],
    children=[
        OptionalVariableInputParam(
            action_func=theirkey_func('input'),
            terminal=True
        )
    ]
)



MainInputParser = InputParser(
    children = [
        BLANK,
        QUIT,
        CMD_HELP,
        ADD_TREE,
        CREATE_TREE,
        SET_SESSION_NAME,
        ACTIVATE_TREE,
        DEACTIVATE_TREE,
        ECHO,
        LIST_ALL_TREES,
        LIST_TREE,
        RENAME_TREE,
        ON_SYSTEM,
        OFF_SYSTEM,
        SET_SAVE_FILE,
        CMD_CD,
        CMD_LS,
        CMD_CDLS,
        CMD_RECROOT,
        CMD_SET_LOCALPATH,
        CMD_REC,
        CMD_REMOVE,
        CMD_SAVE,
        DROP_TREE,
        PRINT_WORKING_TREE,
        CHANGE_TREE,
        TIE_FILES,
        CMD_INFO,
        CMD_HISTORY,
        CMD_TIES,
        SET_HASH,
        SET_OS_TYPE,
        PRINT_WORKING_DIRECTORY,
        UPDATE_ENTRY,
        CMD_NOTES,
        CMD_AUTOUPDATE,
        IGNORE_LIST
    ]
)



cmd_confirmation_prompt = 'confirm_action'


CONFIRMATION_YES = CommandInput(
    'y', 'yes', 'continue', 'ok', 'okay', 'sure', 'yep', 'yeah', 'yup', 'true',
    case_sensitive=False,
    default_data=blank_dict(cmd_confirmation_prompt, 'confirm'),
    final=True,
    terminal=True,
    action_func=mykey_func('confirm', True)
)


CONFIRMATION_NO = CommandInput(
    '', ' ', 'n', 'no', 'cancel', 'stop', 'jk', 'nope', 'nah', 'nevermind', 'false',
    case_sensitive=False,
    default_data=blank_dict(cmd_confirmation_prompt, 'confirm'),
    final=True,
    terminal=True,
    action_func=mykey_func('confirm', False)
)


ConfirmationInputParser = InputParser(
    default_command=CONFIRMATION_NO,
    children=[
        CONFIRMATION_YES,
        CONFIRMATION_NO
    ]
)



"""Message And Command OUT object."""
from copy import deepcopy

class Madout():
    def __init__(self, successvalue=True, message="", *args):
        self.successvalue = successvalue
        self.message = message
        self.data = args

    def __bool__(self):
        return self.successvalue

    def __str__(self):
        return self.message

    def __repr__(self):
        return ("<" + ("Successful" if self else "Failed") +
                " Madout object with text '" + self.message +
                f"' and {len(self.data)} output(s).>")

    def __getitem__(self, item):
        return self.data[item]

    def p(self):
        print(self.message)
        for dto in filter(lambda x: type(x) is Madout, self.data):
            print(dto.message)

    def addtext(self, other):
        cpy = deepcopy(self)
        if type(other) is Madout:
            cpy.message += (' ' + other.message)
        elif type(other) is str:
            cpy.message += (' ' + other)
        return cpy

    def append(self, other):
        self.data = (*self.data, other)

    def result(self):
        if self.data==():
            return ()
        return self.data[0]


class Madgood(Madout):
    def __init__(self):
        super().__init__(True, "")

    def __repr__(self):
        return "<Successful Madgood object.>"


# navtuils - utility functions for nagivating tree files
#   and path manipulation.


from treerec.command_parser import Madout
from treerec.primitives import PARENT
from treerec.pathtuputils import *
from treerec.trace_ancestors import find_current_root
from treerec.dirutils import *


class SuperPath():
    def __init__(self, destination_dict):
        self.name = destination_dict['name']
        self.path = destination_dict['path']

        root = find_current_root(destination_dict)

        self.root_nickname = root.get('nickname')
        self.root_path = root['path']
        self.root_localpath = root.get('localpath', ())
        self.after_root = subtract_paths(root['path'], destination_dict['path'])
        self.localpath = (*self.root_localpath, *self.after_root)

    def print_path(self, *additions):
        return combine_pathtuples(self.path, *additions)

    def print_localpath(self, *additions):
        return combine_pathtuples(self.localpath, *additions)

    def print_root_path(self, *additions):
        return combine_pathtuples(self.root_path, *additions)

    def print_root_localpath(self, *additions):
        return combine_pathtuples(self.root_localpath, *additions)

    def print_after_root(self, *additions):
        return combine_pathtuples(self.after_root, *additions)


def get_localpath_from_path(direc):
    """
    Assuming that the 'path' of direc is a treepath,
    find the localpath using its root's localpath.
    """
    root = find_current_root(direc)
    after_root = subtract_paths(root['path'], direc['path'])
    return (*root['localpath'], *after_root)


def get_path_from_localpath(direc):
    """
    Assuming that 'path' of direc is a localpath,
    find the treepath using its root's treepath.
    """
    root = find_current_root(direc)
    after_root = subtract_paths(root['localpath'], direc['path'])
    return (*root['path'], *after_root)


def cd_virt(cdir:dict, dirname:str) -> Madout:
    """Virtual change directory."""

    if dirname=='.':
        return Madout(True, "", cdir)
    if dirname=='..':
        return Madout(True, "", cdir[PARENT])

    entry, list_from, idx = find_in_dir(cdir, dirname)
    if entry==False:
        return Madout(False, "does not exist")
    if list_from=='file_list':
        return Madout(False, "not a directory")
    if list_from=='dir_list':
        if entry['type']=='unrecorded':
            return Madout(False, "not recorded")
        return Madout(True, "", entry)


def find_best_root(devdict:dict, pathtuple:tuple, use_local:bool=False) -> dict:
    """
    Finds the best root of the path: the ancestor root 
        with the most path parts in common.
    If no roots are found, returns the original directory.
    Expects pathtuple to be an absolute path.
    Works even when devdict is not DEVINFO.
    """
    max_similarity_len = 0
    curdir = devdict
    root_path_to_use = ('localpath' if use_local else 'path')
    for rootdir in devdict['dir_list']:
        if is_ancestor(rootdir.get(root_path_to_use, ()), pathtuple) and (len(rootdir.get(root_path_to_use, ())) > max_similarity_len):
            curdir = rootdir
            max_similarity_len = len(rootdir[root_path_to_use])
    return curdir


def get_dict_from_path(head_dict:dict, pathtuple:tuple, use_local:bool=False) -> Madout:
    """
    Returns the dictionary corresponding to the given path.
    Assumes a full absolute path is given.
    Works even if head_dict is not DEVINFO.
    Accounts for root nicknames, but cannot account for paths starting with '.', '..', or '^'.
    """
    path_key = ('localpath' if use_local else 'path')
    rootdir = find_best_root(head_dict, pathtuple, use_local)
    if rootdir==head_dict:
        return Madout(False, "Root not found")
    idx = len(rootdir[path_key])
    mad_curdir = Madout(True, "", rootdir)
    olddir = rootdir
    while idx < len(pathtuple):
        mad_curdir = cd_virt(olddir, pathtuple[idx])
        if not mad_curdir:
            rootpath = combine_pathtuples(rootdir[path_key])
            afterpath = combine_pathtuples((),
                                           subtract_paths(rootdir['path'], olddir['path']),
                                           (pathtuple[idx],))
            return Madout(False, "File (" + rootpath + ") " + afterpath + " not found in tree.")
        olddir = mad_curdir.result()
        idx += 1
    return mad_curdir


def get_dict_from_path_after_start(start_dict, pathtuple, localpath_output=False):
    """
    Traverses path from start_dict, and returns corresponding dict.
    """
    if len(pathtuple)==0:
        return Madout(True, "", start_dict)
    idx = 0
    mad_curdir = Madout(True, "", start_dict)
    olddir = start_dict
    while idx < len(pathtuple)-1:
        mad_curdir = cd_virt(olddir, pathtuple[idx])
        if not mad_curdir:
            SP = SuperPath(olddir)
            failure_path = (SP.print_localpath((pathtuple[idx],)) if localpath_output else SP.print_path((pathtuple[idx],)))
            if mad_curdir.message=='not a directory':
                return Madout(False, "File " + failure_path + ' is not a directory.')
            if mad_curdir.message=='not recorded':
                return Madout(False, "File " + failure_path + ' is not recorded.')
            if mad_curdir.message=='does not exist':
                return Madout(False, "File " + failure_path + ' does not exist.')
            else:
                return Madout(False, "File " + failure_path + ' not found.')
        olddir = mad_curdir.result()
        idx += 1
    last_entry, list_from, ii = locate_from_dir(olddir, pathtuple[idx])
    if last_entry==False:
        return Madout(False, "File " + combine_pathtuples(olddir['path'], (pathtuple[idx],)) + ' not found.')
    return Madout(True, "", last_entry)



# Path Tuple Utilities - functions for manipulating pathtuples,
# where each entry is a directory/file name.
# Implicity uses pathlib.Path objects, but calls no constructors.


def combine_pathtuples(*pathtuples):
    if len(pathtuples)==0:
        return ''
    if len(pathtuples[0])==0:
        outstr = ['']
    else:
        outstr = ["" if pathtuples[0][0]=='/' else (pathtuples[0][0])]
    outstr.extend(pathtuples[0][1:])
    for pathpiece in pathtuples[1:]:
        outstr.extend(pathpiece)
    if outstr==['']:
        return '/'
    return '/'.join(outstr)


def split_pathstr(pathstr, split_on=['/', '\\']):
    pathlist = [pathstr]
    for split_char in split_on:
        newlist = []
        for entry in pathlist:
            newlist.extend(entry.split(split_char))
        pathlist = newlist
    if pathlist[0]=='':
        pathlist[0] = '/'
    return pathlist


def compare_paths(pathtuple1, pathtuple2):
    """
    Finds how many starting entries in pathtuple1 are identical to those in pathtuple2.
    This function is reflexive.
    pathtuple[idx] is the first element where the paths differ.
    """
    idx = 0
    minlength = min(len(pathtuple1), len(pathtuple2))
    while (idx < minlength) and (pathtuple1[idx] == pathtuple2[idx]):
        idx += 1
    return idx


def subtract_paths(shorttup, longtup):
    if len(shorttup)>len(longtup):
        print("Subtracted path must not be longer than the other path.")
        return False
    idx = compare_paths(shorttup, longtup)
    return longtup[idx:]


def is_ancestor(pathtuple1, pathtuple2):
    """
    Checks if pathtuple1 is an ancestor of pathtuple2.
    This function is not reflexive.
    """
    return (compare_paths(pathtuple1, pathtuple2) == len(pathtuple1))





from copy import copy

# Helper functions for creating action_func functions and initial parameter dictionaries.

def mykey_func(key, value):
    """Returns a function that sets the key of a dict to
    the value given in this file."""
    def setkey(dictin, cval):
        dictin[key] = value
        return copy(dictin)
    return setkey

def set_command(commandname):
    def setkey(dictin, cval):
        dictin['command'] = commandname
        return copy(dictin)
    return setkey

def theirkey_func(key):
    """
    Returns a function that sets the key of a dict to
    the value entered on the command line.
    """
    def setkey(dictin, cval):
        dictin[key] = cval
        return copy(dictin)
    return setkey

def blank_func():
    """
    Returns a function that passes the dict through unchanged.
    """
    def passfunc(dictin, cval):
        return dictin
    return passfunc

def blank_dict(command, *args):
    return {'command': command} | {arg: None for arg in args}


import json, os, time
from pathlib import *
from treerec.primitives import *
from treerec.hist_events import *
from treerec.dictutils import *
from treerec.pathtuputils import *
from treerec.dirutils import *
from treerec.navutils import *
from treerec.entryutils import *


forbidden_root_nicknames = {
    '.', '..', '^', '^^', '*', '/', '\\'
    }


def active(func):
    def must_be_active_wrapper(trobj, *args, **kwargs):
        if trobj.is_active():
            return func(trobj, *args, **kwargs)
        else:
            print('Tree must be active to perform this operation.')
    return must_be_active_wrapper


def onsystem(func):
    def must_be_onsystem_wrapper(trobj, *args, **kwargs):
        if trobj.onsystem:
            return func(trobj, *args, **kwargs)
        else:
            print('Tree must be on its host system to perform this operation.')
    return must_be_onsystem_wrapper


def unsaved(func):
    def set_unsaved_changes_wrapper(trobj, *args, **kwargs):
        out = func(trobj, *args, **kwargs)
        if out:
            trobj.has_unsaved_changes = True
            trobj.dict['mtime'] = time.time()
        return out
    return set_unsaved_changes_wrapper


def ls_dict(dirin:dict, recorded_only:bool=False, show_mtime:bool=False):
    if dirin['type']=='unrecorded':
        print(dirin['name'] + ' is not a recorded directory.')
        return
    if dirin['type'] not in DIR_LIST_TYPES:
        print(dirin['name'] + ' is not a directory.')
        return
    if dirin['type']==DEVINFO:
        for rootentry in dirin['dir_list']:
            rootpath = rootentry.get('path', None)
            if rootpath is None:
                pathprint = 'KEY ERROR: "path"'
            elif rootpath==():
                pathprint = '/'
            else:
                pathprint = combine_pathtuples(rootpath)
            root_local_path = rootentry.get('localpath', None)
            if not root_local_path:
                localprint = 'KEY ERROR: "localpath"'
            elif root_local_path==():
                localprint = '/'
            else:
                localprint = combine_pathtuples(root_local_path)
            print('  <> ' + rootentry.get('nickname', 'ERROR') +
                  '  //  path: (' + pathprint + ')' +
                  '   local: [' + localprint + ']')
        return
    for entry in dirin['dir_list']:
        if not (recorded_only and entry['type']=='unrecorded'):
            print(' ' + 
                  ('x' if entry['type']=='unrecorded' else ' ') +
                  ('*' if len(entry.get('notes', [])) > 0 else ' ') +
                  '[] ' + entry['name'] + 
                  ('     ' + time_formatted(entry['mtime']) 
                    if (show_mtime and entry['type']!='unrecorded')
                     else '')
                )
    for entry in dirin['file_list']:
        if not (recorded_only and entry['type']=='unrecorded'):
            print(' ' + 
                  ('x' if entry['type']=='unrecorded' else ' ') +
                  ('*' if len(entry.get('notes', [])) > 0 else ' ') +
                  '   ' + entry['name'] + 
                  ('     ' + time_formatted(entry['mtime']) 
                    if (show_mtime and entry['type']!='unrecorded')
                    else '')
                )


def print_ties(entry):
    if len(entry['ties'])==0:
        print('>> [ no ties.]')
        return
    for tie in entry['ties']:
        name = tie.get('name', FIELD_NOT_RECORDED)
        device = tie.get('device', FIELD_NOT_RECORDED)
        pathstr = combine_pathtuples(tie.get('path', ()))
        nickname_path = tie.get('nickname_path',())
        root_nickname = ''
        path_after_nickname_str = ''
        tie_time = (time_formatted(tie['time']) if tie.get('time', False) else FIELD_NOT_RECORDED)
        if len(nickname_path)>=1:
            root_nickname = nickname_path[0]
        if len(nickname_path)>=2:
            path_after_nickname_str = combine_pathtuples(nickname_path[1:])
        print( ('* ' if tie.get('identical_hash', False) else '  ') + '>> ' + name + '  <' + device + '>')
        print('   ' + pathstr)
        print('   ' + '(' + root_nickname + ')/' + path_after_nickname_str)
        print('   Session: ' + (tie['session'] if tie['session']!=None else 'None'))
        print('   Time: ' + tie_time)
        print('---------------')
    

def pathtup_without_windows_drive_backslash(pathlike):
    """
    Removes the backslash from the start of a Windows pathlike object.
    """
    if not isinstance(pathlike, PureWindowsPath):
        return pathlike.parts
    drive_name = pathlike.parts[0]
    if len(drive_name)!=0 and drive_name[-1]=='\\':
        drive_pure = drive_name[:-1]
    return (drive_pure, *pathlike.parts[1:])




class TreeObj():
    """
    TreeObj class: fundamental class for holding, manipulating, and displaying tree information.

    Stored info:
        device name
        session name
        device owner name
    
    Tree must be `active` before any changes can be made of any kind,
    including tying files together, renaming, moving, or deleting files.

    Tree must be `onsystem` before any new directories/files can be recorded.

    The tree is stored as a nested series of dictionaries.
    The top-level dictionary is named DEVINFO (defined in dictutils, defaults to 'devinfo'), and has device informaiton.
    DEVINFO contains directories called 'roots', which are meant to correspond
    to different drives on a machine. Each root has it's own 'treepath', which
    is what is used to calculate the paths of each of its children.
        - This is so that the paths have a consistent name, even if the drive
        name is not consistent between sessions.
    """

    def __init__(self, device=None, session=None, onsystem=False, owner=None):
        self.onsystem = onsystem
        self.session = session
        self.device = device
        if onsystem:
            self.set_onsystem()
        self._showname = device
        # self.mPath = self.decide_Path()
        self._ACTIVE = False
        self.hash_files = True  # do we compute SHA256 hash when recording files?
        self.auto_update = True  # do we automatically update the listings of directoies that we cd into?
        self.os_type = None  # This allows us to override the recorded os of entries.
        #  This is helpful when accessing NT drives through Linux OS, for instance.
        self.savefile = None
        self.savefile_active = False
        self.has_unsaved_changes = False
        self.dict = create_dict_pure(
                objname=DEVINFO,
                objtype=DEVINFO,
                objpath=('',),
                owner=owner, 
                device=self.device,
                os_type=os.name,
                ctime=int(time.time()),
                mtime=int(time.time()),
                atime=int(time.time()),
                nickname=DEVINFO,
                include_uuid=True
                )
        self.dict[PARENT] = self.dict
        self.dict['history'].append(record_entry(self.dict, session))
        self.CDIR = self.dict
        self.autoupdate_entries = False
        self.ignore_names_list = []
    

    ### Setting and checking basic properties of the system,
    ### such as `active`, `onsystem`, and `savefile`.


    def set_onsystem(self, device:str=None):
        if device:
            self.set_device(device)
        if not self.device:
            print('Please set device name before asserting that the tree is on the system.')
            return
        self.onsystem = True
        print('Tree on system device "' + self.device + '".')
        print('Auto-update is ' + ('on' if self.auto_update else 'off'))
        print('File hashing is ' + ('on' if self.hash_files else 'off'))
        print('Remember to check the local paths of the tree roots, \n'
              ' as they need to reflect the locations of the files on the system.')

    def set_offsystem(self):
        self.onsystem = False

    def activate(self, sessionname:str=None):
        if sessionname:
            self.session = sessionname
        if self.session==None:
            print('Set session name before activating tree.')
            return
        self._ACTIVE = True

    def deactivate(self):
        self._ACTIVE = False

    def is_active(self):
        return self._ACTIVE
    
    def set_showname(self, name):
        self._showname = name

    def set_device(self, devname:str):
        if not dir_is_empty(self.dict):
            print('Device name can only be changed when tree is first created '
                  '(i.e. when there are no roots: the DEVINFO has no children.)')
            return False
        self.device = devname
        self.dict['device'] = devname
        return True

    @unsaved
    def set_savefile(self, filename=None, quiet=False, force=False):
        # This has the @unsaved decorator, because if the Tree is set to a new savefile,
        # it has not been saved to that savefile yet.
        if filename==None:
            if self.savefile!=None:
                print("Setting load file (" + str(self.savefile) + ") as save file...")
                self.savefile_active = True
                return False
            else:
                print("File name not entered.")
                return False
        if os.path.exists(filename):
            if not quiet: print("Warning: save file already exists.")
            if not force:
                if not quiet: print("Save file not changed.")
                return False
        self.savefile = Path(filename).resolve().absolute()
        self.savefile_active = True
        print("Save file set to " + str(self.savefile))
        return True

    def set_os_type(self, os_type:str=None):
        if os_type not in ['nt', 'posix', None]:
            print('Unusual OS type... Windows = "nt", Linux/Mac = "posix".')
        self.os_type = os_type 

    def print_os_type(self):
        return (self.os_type if self.os_type else os.name)

    def set_hash_rule(self, do_hash:bool):
        self.hash_files = bool(do_hash)

    def make_settings(self):
        return {
                'hash_files': self.hash_files,
                'auto_update': self.auto_update,
                }

    
    ###   Saving and loading


    def to_json(self, settings:dict=dict(), indent=None):
        return json.dumps({'settings': settings, 'data':self.dict}, skipkeys=True, indent=indent)

    @active
    def save(self, pretty:bool=False):
        if not self.savefile_active:
            print("File saving not activated. Please set a save file.")
            return False
        try:
            with open(self.savefile, mode='w') as fopen:
                indent = ('  ' if pretty else None)
                fopen.write(self.to_json(
                    settings=self.make_settings(),
                    indent=indent
                    ))
                self.has_unsaved_changes = False
                return True
        except Exception as error:
            print(error)
            return False

    def load_dict(self, dictin):
        self.dict = setup_parents(dictin)
        self.onsystem = False
        self._ACTIVE = False
        self.device = self.dict['device']
        self._showname = self.device
        self.CDIR = self.dict
        self.dict['atime'] = int(time.time())
        if dictin['os_type'] != os.name:
            print('WARNING: loaded os type not equal to system os type. Double-check for accuracy.')
            self.set_os_type(dictin['os_type'])

    def load_settings(self, settings_dict):
        """
        Load some values from a settings dict into the tree.
        """
        self.hash_files = settings_dict.get('hash_files', True)
        self.auto_update = settings_dict.get('auto_update', True)

    def load_json_file(self, filename, safe=True):
        pathlike = Path(filename).resolve()
        try:
            with open(pathlike, mode='r') as fopen:
                try:
                    file_dict = json.load(fopen)
                except:
                    print("Error loading json from file " + filename)
                    return
        except Exception as error:
            print('Error opening file ' + filename + ':')
            print(error)
            return
        if not file_dict.get('data', False):
            if file_dict['type']==DEVINFO:
                dictin = file_dict
                settings = dict()
            else:
                print('Error loading file: cannot find DEVINFO of json file.')
                return
            print('Old format file loaded...')
        else:
            dictin = file_dict['data']    
            settings = file_dict['settings']
        self.load_dict(dictin)
        self.load_settings(settings)
        self.savefile = pathlike
        return file_dict.get('settings', None)


    ###   Traversing paths: returning the correct dict given a string


    def ret_root_or_cdir_file_from_path_beginning(self, pathparts) -> tuple[dict, tuple, str]:
        """
        Searches the first entries of pathparts to find a good starting directory for path traversal.
        Returns a dict directory, the pathparts that come after it, and the name of
            the child list that the dict is found in.
        First, if the first entry is '.', return self.CDIR.
        Then, it searches to find a root that's an ancestor.
        Then, it searches the current directory for an object.
        Then, it searches the root nicknames.
        If all fails, return False.
        """
        if len(pathparts)==0:
            return False, False, False
        # Search for ancestor root:
        cdir = find_best_root(self.dict, pathparts)
        if cdir!=self.dict:
            path_after_start = subtract_paths(cdir['path'], pathparts)
            return cdir, path_after_start, 'dir_list'
        # Search inside self.CDIR:
        # locate_from_dir also accounts for '..'
        if pathparts[0]=='.':
            return self.CDIR, pathparts[1:], 'dir_list'
        if pathparts[0]=='..':
            return self.CDIR[PARENT], pathparts[1:], 'dir_list'
        nextdir, x_list, idx = find_in_dir(self.CDIR, pathparts[0])
        if nextdir:     # this means that the path is relative, with no './' at the start.
            return self.CDIR, pathparts, 'dir_list'
        # Search for root nickname:
        nextdir, x_list, idx = find_in_dir(self.dict, pathparts[0], nickname_first=True)
        if nextdir:
            return nextdir, pathparts[1:], x_list
        return False, False, False

    def get_path_initial_dict(self, pathstring:str) -> tuple[dict, tuple]:
        """
        Traverses the input pathstring, and returns the most primitive dict of the path,
            e.g., the root directory (for absolute paths or '^') 
            or the current directory (for relative paths, '.').
        Handles '.', '..', '^', root nicknames, absolute (tree) paths, and relative paths.
        """
        pathstring = pathstring.strip()
        pathparts = PurePosixPath(pathstring).parts
        if len(pathparts)==0:
            return self.CDIR, ()
        elif pathparts[0]=='^':
            # This is the character for the current root.
            if self.CDIR is self.dict:
                print('Error: cannot use "current root" identifier (^) when current directory is DEVINFO.')
                return False, ()
            cdir = find_current_root(self.CDIR)
            path_after_start = pathparts[1:]
        elif pathparts[0]=='^^':
            # This is the character combo for devinfo.
            cdir = self.dict
            path_after_start = pathparts[1:]
        else:
            # Consideres '..', '.', root paths, and root nicknames.
            cdir, path_after_start, x_list = self.ret_root_or_cdir_file_from_path_beginning(pathparts)
            if x_list=='file_list':
                if len(path_after_start)!=0:
                    print('Error: ' + pathparts[0] + ' is a file, not a directory.')
                    return False, ()
                else:
                    return cdir, ()
        if not cdir:
            print('Error: could not find starting directory.')
            return False, ()
        return cdir, path_after_start
        # Now, we traverse the path
    
    def traverse_path(self, pathstring:str) -> dict:
        cdir, path_after_start = self.get_path_initial_dict(pathstring)
        if len(path_after_start)==0:
            return cdir
        mad_cdir = get_dict_from_path_after_start(cdir, path_after_start)
        if not mad_cdir:
            print(mad_cdir)
            return False
        return mad_cdir.result()
    
    def get_parent_of_path_file(self, pathstring:str) -> dict:
        cdir, path_after_start = self.get_path_initial_dict(pathstring)
        if not cdir:
            return False
        if len(path_after_start)==0:
            # e.g. path points to '.', '..', a root, etc.
            return cdir.get(PARENT, None)
        mad_cdir = get_dict_from_path_after_start(cdir, path_after_start[:-1])
        if not mad_cdir:
            print(mad_cdir)
            return False
        return mad_cdir.result()
        
    
    ###   Unix-like commands for viewing and changing directories


    def ls(self, dirpath:str=None, recorded_only:bool=False, show_mtime:bool=False):
        if not dirpath:
            dirin = self.CDIR
        else:
            dirin = self.traverse_path(dirpath)
        if not dirin:
            return
        if self.onsystem and self.auto_update and (self.CDIR is not self.dict):
            update_unrecorded_contents(self.CDIR, quiet=True)
        ls_dict(dirin, recorded_only, show_mtime)

    def lsroots(self):
        ls_dict(self.dict)

    def roots(self):
        ls_dict(self.dict)

    def cd_cdir(self, dirname):
        """
        Old version, withou using self.traverse_path
        """
        if self.CDIR['type']==DEVINFO:
            self.chroot(dirname)
            return
        newdir_out = cd_virt(self.CDIR, dirname)
        if not newdir_out:
            if newdir_out.message=="does not exist":
                print(dirname + " does not exist in the current directory.")
            elif newdir_out.message=="not a directory":
                print(dirname + " is not a directory.")
            elif newdir_out.message=="not recorded":
                print(dirname + " must be recorded before it can be accessed.")
            return
        self.CDIR = newdir_out.result()

    def cd(self, dirpath):
        newdir_out = self.traverse_path(dirpath)
        if not newdir_out:
            return False
        if newdir_out['type']=='unrecorded':
            print(dirpath + " must be recorded before it can be accessed.")
            return False
        if newdir_out['type'] not in DIR_LIST_TYPES:
            print(dirpath + " is not a directory.")
            return False
        self.CDIR = newdir_out
        return True

    def chroot(self, rootname):
        for entry in self.dict['dir_list']:
            if entry.get('nickname')==rootname:
                self.CDIR = entry
                return
        print(rootname + " is not the nickname of any root.")
        return

    def pwd(self, full=False):
        SP = SuperPath(self.CDIR)
        print(self._showname + ": (" +
              (SP.print_root_path() if full in ['tree', 'treepath']
               else SP.print_root_localpath() if full in ['local', 'localpath']
               else SP.root_nickname) +
             ")/" + SP.print_after_root())


    ###    Recording new entries


    @onsystem
    @active
    @unsaved
    def recroot(self, root_localpath:str, nickname:str=None, treepath:str=None):
        """
        Create new root entry in the `dirinfo` directory.
        """
        root_localpathlike = Path(root_localpath).resolve()
        if not root_localpathlike.is_dir():
            print('Root directory must be a directory.')
            return False
        if nickname in forbidden_root_nicknames:
            print('Cannot use "' + nickname + '" as a Root Nickname.')
            return False
        if nickname=='':
            print('You cannot have an empty Root Nickname.')
            return False
        if nickname==None:
            print('You should really set a Nickname manually...')
            if root_localpathlike.name in forbidden_root_nicknames:
                print('Cannot use file''s name, "' + root_localpathlike.name +
                      '", to be the Root Nickname. Please choose another Nickname.')
                return False
            nickname = root_localpathlike.name
        if '/' in nickname or '\\' in nickname:
            print('Nickname cannot include "/" or "\\".')
            return False
        current_root_names = [entry['nickname'] for entry in self.dict['dir_list']]
        if nickname in current_root_names:
            print('Root with Nickname "' + nickname + '" already exists.')
            return False
        if treepath==None:
            treepath_tup = pathtup_without_windows_drive_backslash(root_localpathlike)
        elif treepath=='':
            print('Root cannot have an empty Tree path.')
            return False
        elif type(treepath) is tuple:
            treepath_tup = treepath
        else:
            treepath_tup = PurePosixPath(treepath).parts
        if treepath_tup==():
            print('Root cannot have an empty Tree path.')
            return False
        current_root_treepaths = [entry['path'] for entry in self.dict['dir_list']]
        if treepath_tup in current_root_treepaths:
            print('Root with Tree path ' + combine_pathtuples(treepath_tup) + ' already exists.')
            return False
        newentry = create_dict(root_localpathlike, rootname=nickname, nickname=nickname, session=self.session, do_hash=self.hash_files)
        newentry['localpath'] = root_localpathlike.parts
        newentry['path'] = treepath_tup
        newentry[PARENT] = self.dict
        newentry['type'] = 'rootdir'
        self.dict['dir_list'].append(newentry)
        return True

    @active
    @unsaved
    def root_set_treepath(self, rootname:str, recpath:str):
        """
        Set the tree (record) path of a root, i.e. the path which is recorded in the tree to it and its descendants.
        This is static, and can only be set once, when the root has no recorded files.
        Really, this should be set when first recording the tree (using self.recroot())
        """
        if rootname=='^':
            root = find_current_root(self.CDIR)
        else:
            root, _, _ = find_in_dir(self.dict, rootname, nickname_first=True)
        if not root:
            print("Root " + rootname + " not found.")
            return False
        if not dir_is_empty(root):
            print('Root must be empty before setting record path (treepath).')
            return False
        if root['type']==DEVINFO:
            print("Cannot set treepath on DEVINFO, only Roots.")
            return False
        pathparts = PurePosixPath(recpath.strip()).parts
        current_tree_paths = [entry['path'] for entry in self.dict['dir_list']]
        if pathparts in current_tree_paths:
            print('Root with tree path ' + combine_pathtuples(pathparts) + ' already exists.')
            return False
        root['path'] = pathparts
        return True
    
    def root_set_localpath(self, rootname:str, localpath:str):
        """
        Set the local path of a root, i.e. the path on the host system to the root folder.
        We expect the local path of the root to change often, especially for an archival drive,
            which will be connected to many devices in order to record them.
        """
        if rootname=='^':
            root = find_current_root(self.CDIR)
        else:
            root, _, _ = find_in_dir(self.dict, rootname)
        if not root:
            print("Root " + rootname + " not found.")
            return
        if root['type']==DEVINFO:
            print('Cannot set localpath for DEVINFO, only for Roots.')
            return
        pathparts = Path(localpath.strip()).resolve().absolute().parts
        current_local_paths = [entry['localpath'] for entry in self.dict['dir_list']]
        if pathparts in current_local_paths:
            print('Root with local path ' + combine_pathtuples(pathparts) + ' already exists.')
            return
        root['localpath'] = pathparts

    @onsystem
    @active
    @unsaved
    def record(self, objname:str, cdir:dict=None, quiet:bool=False) -> bool:
        """
        Very basic record function. Does not support path entries.
        Should not be called by the user.
        Records <objname> from inside <cdir>.

        Notes: (on Windows)
            - Path(file) does not update st_atime
            - os.stat(file) does note update st_atime
        """
        if cdir==None:
            cdir = self.CDIR
        if cdir['type']=='unrecorded':
            print(cdir['name'] + " must be recorded before its contents can be accessed.")
            return False
        return new_entry_in_dir(
            cdir,
            objname,
            device=self.device,
            os_type=self.os_type,
            session=self.session,
            quiet=quiet,
            do_hash=self.hash_files
            )

    @onsystem
    @active
    def rec(self, pathstr:str, verbose:bool=False, quiet:bool=False) -> bool:
        """
        Path RECord:
        Record file, given absolute path tuple (treepath)
        Uses treepath, not localpath
        """
        parent_dict = self.get_parent_of_path_file(pathstr)
        if not parent_dict:
            return False
        if parent_dict==self.dict:
            print('Cannot record files in DEVINFO. Use recroot to record a new root directory.')
            return False
        objname = PurePosixPath(pathstr).name
        if verbose: print('R:   ' + combine_pathtuples(parent_dict.get('path', ('',)), (objname,)))
        return self.record(objname, parent_dict, quiet=quiet)

    @onsystem
    @active
    def lrec(self, pathin:str, verbose:bool=False, quiet:bool=False) -> bool:
        """
        Local RECord:
        Records file, given absolute pathlike (local path) object.
        We first search through the trees' local paths to find the best root.
        Then, we replace the root's local path in the input path with its tree path,
            traverse through the tree's directory structure, and record the file.
        """
        if type(pathin) is Path:
            pathtuple = pathin.parts
        elif type(pathin) is tuple:
            pathtuple = pathin
        elif type(pathin) is str:
            pathtuple = Path(pathin).resolve().absolute().parts
        else:
            print("Error: path must be a string, Path, or path tuple.")
            return False
        mad_parent_dict = get_dict_from_path(self.dict, pathtuple[0:-1], use_local=True)
        if not mad_parent_dict:
            print(mad_parent_dict)
            return False
        return self.record(pathtuple[-1], mad_parent_dict.result(), quiet=quiet)

    @onsystem
    @active
    def arec(self, cdir:dict=None, verbose:bool=False, head_path:list=[], global_ignore:list=[], quiet:bool=True) -> None:
        """
        All RECord:
        Record all files in directory.
        """
        if cdir==None:
            cdir = self.CDIR
        if cdir['name'] in self.ignore_names_list+global_ignore:
            return
        if verbose: print('R: >>> ' + combine_pathtuples(('*',), subtract_paths(head_path, cdir.get('path', ('',)))))
        for entry in cdir['dir_list']+cdir['file_list']:
            if entry['name'] not in self.ignore_names_list+global_ignore:
                if verbose: print('R:   ' + combine_pathtuples(('*',), subtract_paths(head_path, cdir.get('path', ('',))), (entry['name'],)))
                self.record(entry['name'], cdir, quiet=quiet)

    @onsystem
    @active
    def rrec(self, dirpath:str='.', levels:int=0, global_ignore:list=[], verbose_levels:int=0, quiet:bool=True) -> bool:
        """
        Recursive RECord.
        levels=0 is identical to self.rec [FILE].
        levels=1 is identical to self.arec [DIR].
        levels<0 will record all possible subdirectories.
        """
        cdir = self.traverse_path(dirpath)
        if not cdir: return False
        if cdir['type']==DEVINFO:
            print("Cannot execute recursive record on DEVINFO; must choose root (or regular) directory.")
            return False
        if cdir['type']=='unrecorded':
            self.rec(dirpath, quiet=quiet)
            cdir = self.traverse_path(dirpath)
        if verbose_levels: 
            print('R:   ' + combine_pathtuples(cdir.get('path', ('',))))
            verbose_levels -= 1
        if levels==0: return False
        if cdir['type'] not in DIR_LIST_TYPES:
            print("Cannot execute recursive record on a file; must be a recorded directory.")
            return False
        self.recursive_record(cdir, levels=levels-1, global_ignore=global_ignore, verbose_levels=verbose_levels, head_path=cdir['path'], quiet=quiet)
        return True

    def recursive_record(self, cdir, levels=0, verbose_levels:int=0, global_ignore:list=[], head_path:list=[], quiet=True):
        if verbose_levels:
            verbose = True
            verbose_levels -= 1
        else:
            verbose = False
        self.arec(cdir, global_ignore=global_ignore, verbose=verbose, head_path=head_path, quiet=quiet)
        if levels != 0:
            for entry in cdir['dir_list']:
                if entry['name'] not in self.ignore_names_list+global_ignore:
                    self.recursive_record(entry, levels=levels - 1, global_ignore=global_ignore, verbose_levels=verbose_levels, head_path=head_path, quiet=quiet)

    @active
    @unsaved
    def remove(self, objpath:str='.', quite:bool=False):
        """
        Removes a recorded file/directory.
        """
        entry = self.traverse_path(objpath)
        if not entry:
            return False
        if entry['type']==DEVINFO:
            print('Cannot remove DEVINFO.')
            return False
        if entry['type']=='rootdir':
            idx, x_list = get_position_in_parent(entry, nickname_first=True)
            if not x_list:
                return False
            parent = entry[PARENT]
            parent[x_list].pop(idx)
            return True
        else:
            idx, x_list = get_position_in_parent(entry, nickname_first=True)
            if not x_list:
                return False
            unrecorded = {
                'name': entry['name'],
                'type': 'unrecorded'
            }
            parent = entry[PARENT]
            parent[x_list][idx] = unrecorded
            return True
    
    ### Ignorelist manipulation
        
    def ignorelist_print(self):
        for idx, name in enumerate(self.ignore_names_list):
            print('[', idx, ']' + '  ' + name)
    
    def ignorelist_remove(self, name:str=None, idx:int=None):
        if name:
            for ii, entry in enumerate(self.ignore_names_list):
                if entry==name:
                    self.ignore_names_list.pop(ii)
                    return True
            print('Name "' + name + '" not found in Ignore List.')
            return False
        elif idx is not None:
            if idx >= len(self.ignore_names_list):
                print('Error: index (' + idx + ') too large for Ignore List.')
                return False
            self.ignore_names_list.pop(idx)
            return True
        else:
            print('Invalid input for ignorelist_remove')
            return False
    
    def ignorelist_add(self, name:str=None):
        if not name:
            print('Ignore List name cannot be empty.')
            return False
        self.ignore_names_list.append(name)
        return True


    ### Update files
        

    @onsystem
    def update(self, objpath:str='.', quiet:bool=False):
        entry = self.traverse_path(objpath)
        if not entry:
            return
        if entry['type']==DEVINFO:
            print('Cannot update DEVINFO: no corresponding system file.')
            return
        _, updated = update_object(
                entry, 
                session=self.session,
                do_hash=self.hash_files,
                quiet=quiet)
        if updated:
            self.has_unsaved_changes = True

    @onsystem
    def update_contents(self, objpath:str='.', quiet:bool=False):
        dir_dict = self.traverse_path(objpath)
        if dir_dict['type']==DEVINFO:
            print('Cannot update contents of DEVINFO.')
            return
        update_unrecorded_contents(dir_dict, quiet=quiet)

    @onsystem
    @active
    def r_update(self, objpath:str='.', levels:int=0):
        """
        Recursive update: update folder and all children,
        up to a certain number of levels.
        r_update('.', levels=0) updates all entries in directory.
        """
        cdir = self.traverse_path(objpath)
        self.update(objpath)
        updated = self.recursive_update(cdir, levels)
        if updated:
            self.has_unsaved_changes = True

    def recursive_update(self, cdir, levels:int=0):
        if cdir['type']=='unrecorded':
            return False
        updated = False
        for entry in cdir['dir_list']:
            _, tempu = update_object(entry, session=self.session)
            updated = any([updated, tempu])
        for entry in cdir['file_list']:
            _, tempu = update_object(entry, session=self.session)
            updated = any([updated, tempu])
        if levels==0:
            return
        for direc in cdir.get('dir_list', []):
            tempu = self.recursive_update(direc, levels=levels-1)
            updated = any([updated, tempu])
        return updated

    @active
    @unsaved
    def add_note(self, objpath, note_text):
        entry = self.traverse_path(objpath)
        if not entry:
            return False
        return dict_add_note(entry, note_text, session=self.session)

    @active
    @unsaved
    def remove_note(self, objpath, note_idx):
        entry = self.traverse_path(objpath)
        if not entry:
            return False
        try:
            entry['notes'].pop(note_idx)
            return True
        except IndexError:
            print("Note # " + str(note_idx) + " out of range.")
            return False


    ### Manipulate present data
        
    @active
    @unsaved
    def move(self, sourcepath:str, destinationpath:str):
        pass


    ### Print information about entries


    def info(self, objpath, option='full'):
        entry = self.traverse_path(objpath)
        if not entry:
            return
        print()
        if option=='full':
            print_dict_info(entry)
        if option=='short':
            print_dict_info(entry,
                            ('nicknamepath','ctime','mtime','atime','abbrev-hash',
                             'notes','ties','history','lastmodifiedsession'))

    def history(self, objpath:str, full=False):
        entry = self.traverse_path(objpath)
        if not entry:
            return
        if not full:
            print()
            print_dict_history(entry)
        else:
            print()
            print_dict_history_full(entry)

    def notes(self, objpath):
        entry = self.traverse_path(objpath)
        if not entry:
            return
        print()
        print_dict_notes(entry)

    def ties(self, objpath:str):
        entry = self.traverse_path(objpath)
        if not entry:
            return
        print()
        print_ties(entry)


if __name__=='__main__':
    T = TreeObj(device='Test', onsystem=True, owner='Nerneg', session="Test1")
    T.activate()
    T.recroot(Path('..').resolve(), 'rootdir')
    T.rrec(levels=5, quiet=False)
    print()



# File for declaring primitive values for treerec.

VERSION_NUMBER = '1.0.3'


# ('parent',) is used as the dictionary key for a dict's parent,
# mostly so that it can be automatically removed when converting the
# tree into json by using json.dumps(skipkeys=True).
# If we did not remove them, then we would have infinite recursion.

PARENT = ('parent',)


# DEVINFO is the name, nickname, and type of the top-level dictionary,
# which contains all of the tree children, along with device info.

DEVINFO = 'DEVINFO'


# these are the types of objects which go under the 'dir_list' and
# 'file_list' lists of directory entries, respectively.

DIR_LIST_TYPES = [DEVINFO, 'rootdir', 'dir', 'shortcut']
FILE_LIST_TYPES = ['file']


# This is the default character to represent the system

DEFAULT_SYSCHAR = '$'


# This is the default entry for a printed field when
# the field does not exist.

FIELD_NOT_RECORDED = '---'


# This is the default entry for any printed hash field
# when the file in question is a directory

NO_HASH_BC_DIR = '-(dir)-'


# This is the default entry for a file where the hash
# is not present due to a lack of file permissions

NO_HASH_BC_PERMISSIONS = '-(denied)-'


# Format string for printing dates & times

DATE_FORMAT = '%d/%b/%Y'

TIME_FORMAT = '%I:%M %p'


from treerec.engine.engine_base import TreeEngine_Base
from pathlib import *
import json


def touch_file(filepath:str) -> bool:
    try:
        with open(filepath,'a') as file:
            file
        return True
    except Exception as error:
        print("Error with file: ", filepath, " : ", error)
        raise error


def new_update_file(pathlike:Path):
    if pathlike.exists():
        print("ERROR: File already exists.")
        return None
    try:
        with open(pathlike, 'w') as file:
            file.write('{}')
    except Exception as error:
        print("ERROR in making new update file: ", error)
        return None
    update_data = load_update_file(pathlike)
    return update_data


def load_update_file(pathlike:Path):
    if not pathlike.exists():
        print("ERROR: File does not exist.")
        return None
    try:
        with open(pathlike, 'r') as file:
            update_data = json.load(file)
    except Exception as error:
        print("ERROR in reading new update file: ", error)
        return None
    return update_data


class TreeEngine_Pending(TreeEngine_Base):

    # Extends MultiTreeHolder to include the use of an update file,
    # Which records events that may affect other trees that are not loaded.
    # E.g., moving a file that is tied to a file in an unloaded tree - 
    #   the file in the unloaded tree needs it's ties changed.
    
    
    # The update data is a dict, where the keys are the UUIDs of the trees.
    # Each element of the dict is a list of updates. Each update is a histevent,
    #   which will be interpreted by the MultiTreeHolder when the tree is loaded.

    def __init__(self, update_file_path:str, session:str=None, new_multitree:bool=False):
        super().__init__(session)
        self.update_data = None
        self.update_file_pathlike = Path(update_file_path).resolve().absolute()
        if new_multitree:
            self.update_data = new_update_file(self.update_file_pathlike)
        else:
            self.update_data = load_update_file(self.update_file_pathlike)
        
    def set_update_file(self, filename:str):
        if self.update_data is not None:
            print("ERROR: cannot set new update file when there is already update data loaded.")
            return
        pathlike = Path(filename).resolve().absolute()
        self.update_data = load_update_file(pathlike)
    
    def set_new_update_file(self, filename:str):
        if self.update_data is not None:
            print("ERROR: cannot create and set new update file when there is already update data loaded.")
            return
        pathlike = Path(filename).resolve().absolute()
        self.update_data = new_update_file(pathlike)
        
    
    ### adding events to the update data.
    ### relevant functions are: move, rename, delete
    
    
                
            
